import logging
import os
from collections.abc import Mapping, Sequence
from enum import Enum
from typing import TYPE_CHECKING, Any, Optional, cast

from dagster import (
    Array,
    Bool,
    String,
    _check as check,
)
from dagster._config import (
    Field,
    IntSource,
    Noneable,
    Permissive,
    ScalarUnion,
    Selector,
    Shape,
    StringSource,
    validate_config,
)
from dagster._config.source import BoolSource
from dagster._core.errors import DagsterInvalidConfigError
from dagster._core.storage.config import mysql_config, pg_config
from dagster._record import record
from dagster._serdes import class_from_code_pointer
from dagster._utils.concurrency import get_max_concurrency_limit_value
from dagster._utils.merger import merge_dicts
from dagster._utils.yaml_utils import load_yaml_from_globs

if TYPE_CHECKING:
    from dagster._core.definitions.run_request import InstigatorType
    from dagster._core.instance import DagsterInstance
    from dagster._core.run_coordinator.queued_run_coordinator import RunQueueConfig
    from dagster._core.scheduler.instigation import TickStatus

DAGSTER_CONFIG_YAML_FILENAME = "dagster.yaml"


def is_dagster_home_set() -> bool:
    return bool(os.getenv("DAGSTER_HOME"))


def dagster_instance_config(
    base_dir: str,
    config_filename: str = DAGSTER_CONFIG_YAML_FILENAME,
    overrides: Optional[Mapping[str, object]] = None,
) -> tuple[Mapping[str, Any], Optional[type["DagsterInstance"]]]:
    check.str_param(base_dir, "base_dir")
    check.invariant(os.path.isdir(base_dir), "base_dir should be a directory")
    overrides = check.opt_mapping_param(overrides, "overrides")

    config_yaml_path = os.path.join(base_dir, config_filename)

    if not os.path.exists(config_yaml_path) and is_dagster_home_set():
        logging.getLogger("dagster").warning(
            f"No dagster instance configuration file ({config_filename}) found at {base_dir}."
            f" Defaulting to loading and storing all metadata with {base_dir}. If this is the"
            f" desired behavior, create an empty {config_filename} file in {base_dir}."
        )

    dagster_config_dict = merge_dicts(load_yaml_from_globs(config_yaml_path) or {}, overrides or {})

    if "instance_class" in dagster_config_dict:
        custom_instance_class_data = dagster_config_dict["instance_class"]

        validate_custom_config = validate_config(
            configurable_class_schema(),
            custom_instance_class_data,
        )
        if not validate_custom_config.success:
            raise DagsterInvalidConfigError(
                f"Errors whilst loading dagster custom class config at {config_filename}",
                validate_custom_config.errors,
                custom_instance_class_data,
            )

        custom_instance_class = cast(
            type["DagsterInstance"],
            class_from_code_pointer(
                custom_instance_class_data["module"],
                custom_instance_class_data["class"],
            ),
        )

        schema: Mapping[str, Field]
        if hasattr(custom_instance_class, "config_schema"):
            schema = merge_dicts(
                dagster_instance_config_schema(),
                custom_instance_class.config_schema(),  # type: ignore
            )
        else:
            schema = dagster_instance_config_schema()
    else:
        custom_instance_class = None
        schema = dagster_instance_config_schema()

    if "run_queue" in dagster_config_dict and "run_coordinator" in dagster_config_dict:
        raise DagsterInvalidConfigError(
            "Found config for `run_queue` which is incompatible with `run_coordinator` config"
            " entry.",
            [],
            None,
        )

    if "storage" in dagster_config_dict and (
        "run_storage" in dagster_config_dict
        or "event_log_storage" in dagster_config_dict
        or "schedule_storage" in dagster_config_dict
    ):
        raise DagsterInvalidConfigError(
            "Found config for `storage` which is incompatible with `run_storage`, "
            "`event_log_storage`, and `schedule_storage` config entries.",
            [],
            None,
        )
    elif "storage" in dagster_config_dict:
        if len(dagster_config_dict["storage"]) != 1:
            raise DagsterInvalidConfigError(
                f"Errors whilst loading dagster storage at {config_filename}, Expected one of:"
                "['postgres', 'mysql', 'sqlite', 'custom']",
                [],
                dagster_config_dict["storage"],
            )

    # validate default op concurrency limits
    if "concurrency" in dagster_config_dict:
        validate_concurrency_config(dagster_config_dict)

    dagster_config = validate_config(schema, dagster_config_dict)
    if not dagster_config.success:
        raise DagsterInvalidConfigError(
            f"Errors whilst loading dagster instance config at {config_filename}.",
            dagster_config.errors,
            dagster_config_dict,
        )

    return (check.not_none(dagster_config.value), custom_instance_class)


def config_field_for_configurable_class() -> Field:
    return Field(configurable_class_schema(), is_required=False)


def run_queue_config_schema() -> Field:
    from dagster._core.run_coordinator.queued_run_coordinator import QueuedRunCoordinator

    return Field(
        QueuedRunCoordinator.config_type(),
        is_required=False,
    )


def validate_concurrency_config(dagster_config_dict: Mapping[str, Any]):
    concurrency_config = dagster_config_dict["concurrency"]
    if "pools" in concurrency_config:
        if concurrency_config.get("default_op_concurrency_limit") is not None:
            raise DagsterInvalidConfigError(
                "Found config for `default_op_concurrency_limit` which is incompatible with `pools` config.  Use `pools > default_limit` instead.",
                [],
                None,
            )

        default_concurrency_limit = check.opt_inst(
            pluck_config_value(concurrency_config, ["pools", "default_limit"]), int
        )
        if default_concurrency_limit is not None:
            max_limit = get_max_concurrency_limit_value()
            if default_concurrency_limit < 0 or default_concurrency_limit > max_limit:
                raise DagsterInvalidConfigError(
                    f"Found value `{default_concurrency_limit}` for `pools > default_limit`, "
                    f"Expected value between 0-{max_limit}.",
                    [],
                    None,
                )

        granularity = concurrency_config.get("pools").get("granularity")
        if granularity and granularity not in ["run", "op"]:
            raise DagsterInvalidConfigError(
                f"Found value `{granularity}` for `granularity`, Expected value 'run' or 'op'.",
                [],
                None,
            )

    elif "default_op_concurrency_limit" in concurrency_config:
        default_concurrency_limit = check.opt_inst(
            pluck_config_value(concurrency_config, ["default_op_concurrency_limit"]),
            int,
        )
        if default_concurrency_limit is not None:
            max_limit = get_max_concurrency_limit_value()
            if default_concurrency_limit < 0 or default_concurrency_limit > max_limit:
                raise DagsterInvalidConfigError(
                    f"Found value `{default_concurrency_limit}` for `default_op_concurrency_limit`, "
                    f"Expected value between 0-{max_limit}.",
                    [],
                    None,
                )

    using_concurrency_config = "runs" in concurrency_config or "pools" in concurrency_config
    if using_concurrency_config:
        conflicting_run_queue_fields = [
            ["max_concurrent_runs"],
            ["tag_concurrency_limits"],
            ["block_op_concurrency_limited_runs", "op_concurrency_slot_buffer"],
        ]
        if "run_queue" in dagster_config_dict:
            for field in conflicting_run_queue_fields:
                if pluck_config_value(dagster_config_dict, ["run_queue", *field]) is not None:
                    raise DagsterInvalidConfigError(
                        f"Found config value for `{field}` in `run_queue` which is incompatible with the `concurrency > runs` config",
                        [],
                        None,
                    )

        if "run_coordinator" in dagster_config_dict:
            if (
                pluck_config_value(dagster_config_dict, ["run_coordinator", "class"])
                == "QueuedRunCoordinator"
            ):
                for field in conflicting_run_queue_fields:
                    if (
                        pluck_config_value(
                            dagster_config_dict, ["run_coordinator", "config", *field]
                        )
                        is not None
                    ):
                        raise DagsterInvalidConfigError(
                            f"Found config value for `{field}` in `run_coordinator` which is incompatible with the `concurrency > runs` config",
                            [],
                            None,
                        )


def pluck_config_value(config: Mapping[str, Any], path: Sequence[str]):
    value = config
    for part in path:
        if not isinstance(value, dict):
            return None

        value = value.get(part)
        if value is None:
            return value

    return value


def storage_config_schema() -> Field:
    return Field(
        Selector(
            {
                "postgres": Field(pg_config()),
                "mysql": Field(mysql_config()),
                "sqlite": Field({"base_dir": StringSource}),
                "custom": Field(configurable_class_schema()),
            }
        ),
        is_required=False,
    )


def configurable_class_schema() -> Mapping[str, Any]:
    return {"module": str, "class": str, "config": Field(Permissive())}


def python_logs_config_schema() -> Field:
    return Field(
        {
            "managed_python_loggers": Field(Array(str), is_required=False),
            "python_log_level": Field(str, is_required=False),
            "dagster_handler_config": Field(
                {
                    "handlers": Field(dict, is_required=False),
                    "formatters": Field(dict, is_required=False),
                    "filters": Field(dict, is_required=False),
                },
                is_required=False,
            ),
        },
        is_required=False,
    )


DEFAULT_LOCAL_CODE_SERVER_STARTUP_TIMEOUT = 180


def get_default_tick_retention_settings(
    instigator_type: "InstigatorType",
) -> Mapping["TickStatus", int]:
    from dagster._core.definitions.run_request import InstigatorType
    from dagster._core.scheduler.instigation import TickStatus

    if instigator_type == InstigatorType.SCHEDULE:
        return {
            TickStatus.STARTED: -1,
            TickStatus.SKIPPED: -1,
            TickStatus.SUCCESS: -1,
            TickStatus.FAILURE: -1,
        }
    # for sensor / auto-materialize
    return {
        TickStatus.STARTED: -1,
        TickStatus.SKIPPED: 7,
        TickStatus.SUCCESS: -1,
        TickStatus.FAILURE: -1,
    }


def _tick_retention_config_schema() -> Field:
    return Field(
        {
            "purge_after_days": ScalarUnion(
                scalar_type=int,
                non_scalar_schema={
                    "skipped": Field(int, is_required=False),
                    "success": Field(int, is_required=False),
                    "failure": Field(int, is_required=False),
                    "started": Field(int, is_required=False),
                },
            )
        },
        is_required=False,
    )


def retention_config_schema() -> Field:
    return Field(
        {
            "schedule": _tick_retention_config_schema(),
            "sensor": _tick_retention_config_schema(),
            "auto_materialize": _tick_retention_config_schema(),
        },
        is_required=False,
    )


def get_tick_retention_settings(
    settings: Optional[Mapping[str, Any]],
    default_retention_settings: Mapping["TickStatus", int],
) -> Mapping["TickStatus", int]:
    if not settings or not settings.get("purge_after_days"):
        return default_retention_settings

    purge_value = settings["purge_after_days"]
    if isinstance(purge_value, int):
        # set a number of days retention value for all tick types
        return {status: purge_value for status, _ in default_retention_settings.items()}

    elif isinstance(purge_value, dict):
        return {
            # override the number of days retention value for tick types that are specified
            status: purge_value.get(status.value.lower(), default_value)
            for status, default_value in default_retention_settings.items()
        }
    else:
        return default_retention_settings


def backfills_daemon_config() -> Field:
    return Field(
        {
            "use_threads": Field(Bool, is_required=False, default_value=False),
            "num_workers": Field(
                int,
                is_required=False,
                description="How many threads to use to process multiple backfills in parallel",
            ),
        },
        is_required=False,
    )


def sensors_daemon_config() -> Field:
    return Field(
        {
            "use_threads": Field(Bool, is_required=False, default_value=False),
            "num_workers": Field(
                int,
                is_required=False,
                description=(
                    "How many threads to use to process ticks from multiple sensors in parallel"
                ),
            ),
            "num_submit_workers": Field(
                int,
                is_required=False,
                description=(
                    "How many threads to use to submit runs from sensor ticks. Can be used to"
                    " decrease latency when a sensor emits multiple run requests within a single"
                    " tick."
                ),
            ),
        },
        is_required=False,
    )


def schedules_daemon_config() -> Field:
    return Field(
        {
            "use_threads": Field(Bool, is_required=False, default_value=False),
            "num_workers": Field(
                int,
                is_required=False,
                description=(
                    "How many threads to use to process ticks from multiple schedules in parallel"
                ),
            ),
            "num_submit_workers": Field(
                int,
                is_required=False,
                description=(
                    "How many threads to use to submit runs from schedule ticks. Can be used to"
                    " decrease latency when a schedule emits multiple run requests within a single"
                    " tick."
                ),
            ),
        },
        is_required=False,
    )


def secrets_loader_config_schema() -> Field:
    return Field(
        Selector(
            {
                # Space to add additional built-in secrets loaders in the future, for now
                # only custom
                "custom": Field(configurable_class_schema()),
            }
        ),
        is_required=False,
    )


def get_concurrency_config() -> Field:
    return Field(
        {
            "pools": Field(
                {
                    "default_limit": Field(
                        int,
                        is_required=False,
                        description="The default maximum number of concurrent operations for an unconfigured pool",
                    ),
                    "granularity": Field(
                        str,
                        is_required=False,
                        description="The granularity of the concurrency enforcement of the pool.  One of `run` or `op`.",
                    ),
                    "op_granularity_run_buffer": Field(
                        int,
                        is_required=False,
                        description=(
                            "When the pool scope is set to `op`, this determines the number of runs "
                            "that can be launched with all of its steps blocked waiting for pool slots "
                            "to be freed."
                        ),
                    ),
                },
                is_required=False,
            ),
            "runs": Field(
                {
                    "max_concurrent_runs": Field(
                        int,
                        is_required=False,
                        description=(
                            "The maximum number of runs that are allowed to be in progress at once."
                            " Defaults to 10. Set to -1 to disable the limit. Set to 0 to stop any runs"
                            " from launching. Any other negative values are disallowed."
                        ),
                    ),
                    "tag_concurrency_limits": Field(
                        config=Noneable(
                            Array(
                                Shape(
                                    {
                                        "key": String,
                                        "value": Field(
                                            ScalarUnion(
                                                scalar_type=String,
                                                non_scalar_schema=Shape(
                                                    {"applyLimitPerUniqueValue": Bool}
                                                ),
                                            ),
                                            is_required=False,
                                        ),
                                        "limit": Field(int),
                                    }
                                )
                            )
                        ),
                        is_required=False,
                        description=(
                            "A set of limits that are applied to runs with particular tags. If a value is"
                            " set, the limit is applied to only that key-value pair. If no value is set,"
                            " the limit is applied across all values of that key. If the value is set to a"
                            " dict with `applyLimitPerUniqueValue: true`, the limit will apply to the"
                            " number of unique values for that key."
                        ),
                    ),
                },
                is_required=False,
            ),
            "default_op_concurrency_limit": Field(
                int,
                is_required=False,
                description="[Deprecated] The default maximum number of concurrent operations for an unconfigured concurrency key",
            ),
        }
    )


def dagster_instance_config_schema() -> Mapping[str, Field]:
    return {
        "local_artifact_storage": config_field_for_configurable_class(),
        "compute_logs": config_field_for_configurable_class(),
        "storage": storage_config_schema(),
        "run_queue": run_queue_config_schema(),
        "run_storage": config_field_for_configurable_class(),
        "event_log_storage": config_field_for_configurable_class(),
        "schedule_storage": config_field_for_configurable_class(),
        "scheduler": config_field_for_configurable_class(),
        "run_coordinator": config_field_for_configurable_class(),
        "run_launcher": config_field_for_configurable_class(),
        "telemetry": Field(
            {"enabled": Field(Bool, is_required=False)},
        ),
        "nux": Field(
            {"enabled": Field(Bool, is_required=False)},
        ),
        "instance_class": config_field_for_configurable_class(),
        "python_logs": python_logs_config_schema(),
        "run_monitoring": Field(
            {
                "enabled": Field(Bool, is_required=False),
                "start_timeout_seconds": Field(int, is_required=False),
                "cancel_timeout_seconds": Field(int, is_required=False),
                "max_runtime_seconds": Field(int, is_required=False),
                "max_resume_run_attempts": Field(int, is_required=False),
                "poll_interval_seconds": Field(int, is_required=False),
                "cancellation_thread_poll_interval_seconds": Field(int, is_required=False),
                "free_slots_after_run_end_seconds": Field(int, is_required=False),
            },
        ),
        "run_retries": Field(
            {
                "enabled": Field(bool, is_required=False, default_value=False),
                "max_retries": Field(int, is_required=False, default_value=0),
                "retry_on_asset_or_op_failure": Field(
                    bool,
                    is_required=False,
                    default_value=True,
                    description="Whether to retry runs that failed due to assets or ops in the run failing. "
                    "Set this to false if you only want to retry failures that occur "
                    "due to the run worker crashing or unexpectedly terminating, and instead "
                    "rely on op or asset-level retry policies to retry asset or op failures. Setting this "
                    "field to false will only change retry behavior for runs on dagster "
                    "version 1.6.7 or greater.",
                ),
            }
        ),
        "code_servers": Field(
            {
                "local_startup_timeout": Field(int, is_required=False),
                "reload_timeout": Field(int, is_required=False),
                "wait_for_local_processes_on_shutdown": Field(bool, is_required=False),
            },
            is_required=False,
        ),
        "secrets": secrets_loader_config_schema(),
        "retention": retention_config_schema(),
        "backfills": backfills_daemon_config(),
        "sensors": sensors_daemon_config(),
        "schedules": schedules_daemon_config(),
        "auto_materialize": Field(
            {
                "enabled": Field(BoolSource, is_required=False),
                "minimum_interval_seconds": Field(IntSource, is_required=False),
                "run_tags": Field(dict, is_required=False),
                "respect_materialization_data_versions": Field(BoolSource, is_required=False),
                "max_tick_retries": Field(
                    IntSource,
                    default_value=3,
                    is_required=False,
                    description=(
                        "For each auto-materialize tick that raises an error, how many times to retry that tick"
                    ),
                ),
                "use_sensors": Field(BoolSource, is_required=False),
                "use_threads": Field(Bool, is_required=False, default_value=False),
                "num_workers": Field(
                    int,
                    is_required=False,
                    description=(
                        "How many threads to use to process ticks from multiple automation policy sensors in parallel"
                    ),
                ),
            }
        ),
        "concurrency": get_concurrency_config(),
    }


class PoolGranularity(Enum):
    OP = "op"
    RUN = "run"


@record
class PoolConfig:
    pool_granularity: Optional[PoolGranularity]
    default_pool_limit: Optional[int]
    op_granularity_run_buffer: Optional[int]


@record
class ConcurrencyConfig:
    pool_config: PoolConfig
    run_queue_config: Optional["RunQueueConfig"]

    @staticmethod
    def from_concurrency_settings(
        concurrency_settings: Mapping[str, Any],
        run_coordinator_run_queue_config: Optional["RunQueueConfig"],
    ) -> "ConcurrencyConfig":
        from dagster._core.run_coordinator.queued_run_coordinator import RunQueueConfig

        run_settings = concurrency_settings.get("runs", {})
        pool_settings = concurrency_settings.get("pools", {})
        pool_granularity_str = pool_settings.get("granularity")
        if run_coordinator_run_queue_config:
            run_queue_config = RunQueueConfig(
                max_concurrent_runs=run_settings.get(
                    "max_concurrent_runs", run_coordinator_run_queue_config.max_concurrent_runs
                ),
                tag_concurrency_limits=run_settings.get(
                    "tag_concurrency_limits",
                    run_coordinator_run_queue_config.tag_concurrency_limits,
                ),
                max_user_code_failure_retries=run_coordinator_run_queue_config.max_user_code_failure_retries,
                user_code_failure_retry_delay=run_coordinator_run_queue_config.user_code_failure_retry_delay,
                should_block_op_concurrency_limited_runs=run_coordinator_run_queue_config.should_block_op_concurrency_limited_runs,
                op_concurrency_slot_buffer=pool_settings.get(
                    "op_granularity_run_buffer",
                    run_coordinator_run_queue_config.op_concurrency_slot_buffer,
                ),
            )
        else:
            run_queue_config = None

        pool_granularity = PoolGranularity(pool_granularity_str) if pool_granularity_str else None
        if (
            not pool_granularity
            and run_coordinator_run_queue_config
            and run_coordinator_run_queue_config.explicitly_enabled_concurrency_run_blocking
        ):
            # if this was explicitly configured in the run coordinator config, we should default to op granularity
            pool_granularity = PoolGranularity.OP

        default_limit = pool_settings.get("default_limit")
        if default_limit is None:
            default_limit = concurrency_settings.get("default_op_concurrency_limit")

        return ConcurrencyConfig(
            pool_config=PoolConfig(
                pool_granularity=pool_granularity,
                default_pool_limit=default_limit,
                op_granularity_run_buffer=pool_settings.get("op_granularity_run_buffer"),
            ),
            run_queue_config=run_queue_config,
        )



from typing import TYPE_CHECKING, Any, Optional

import dagster._check as check
from dagster._core.errors import DagsterUserCodeProcessError
from dagster._core.types.loadable_target_origin import LoadableTargetOrigin
from dagster._serdes import deserialize_value
from dagster._utils.error import SerializableErrorInfo

if TYPE_CHECKING:
    from dagster._grpc.client import DagsterGrpcClient
    from dagster._grpc.types import ListRepositoriesResponse


def sync_list_repositories_grpc(api_client: "DagsterGrpcClient") -> "ListRepositoriesResponse":
    from dagster._grpc.client import DagsterGrpcClient
    from dagster._grpc.types import ListRepositoriesResponse

    check.inst_param(api_client, "api_client", DagsterGrpcClient)
    result = deserialize_value(
        api_client.list_repositories(),
        (ListRepositoriesResponse, SerializableErrorInfo),
    )
    if isinstance(result, SerializableErrorInfo):
        raise DagsterUserCodeProcessError(
            result.to_string(), user_code_process_error_infos=[result]
        )
    else:
        return result


async def gen_list_repositories_grpc(
    api_client: "DagsterGrpcClient", **kwargs: Any
) -> "ListRepositoriesResponse":
    from dagster._grpc.client import DagsterGrpcClient
    from dagster._grpc.types import ListRepositoriesResponse

    check.inst_param(api_client, "api_client", DagsterGrpcClient)
    result = deserialize_value(
        await api_client.gen_list_repositories(**kwargs),
        (ListRepositoriesResponse, SerializableErrorInfo),
    )
    if isinstance(result, SerializableErrorInfo):
        raise DagsterUserCodeProcessError(
            result.to_string(), user_code_process_error_infos=[result]
        )
    else:
        return result


def sync_list_repositories_ephemeral_grpc(
    executable_path: str,
    python_file: Optional[str],
    module_name: Optional[str],
    working_directory: Optional[str],
    attribute: Optional[str],
    package_name: Optional[str],
) -> "ListRepositoriesResponse":
    from dagster._grpc.client import ephemeral_grpc_api_client

    check.str_param(executable_path, "executable_path")
    check.opt_str_param(python_file, "python_file")
    check.opt_str_param(module_name, "module_name")
    check.opt_str_param(working_directory, "working_directory")
    check.opt_str_param(attribute, "attribute")
    check.opt_str_param(package_name, "package_name")

    with ephemeral_grpc_api_client(
        loadable_target_origin=LoadableTargetOrigin(
            executable_path=executable_path,
            module_name=module_name,
            python_file=python_file,
            working_directory=working_directory,
            attribute=attribute,
            package_name=package_name,
        )
    ) as api_client:
        return sync_list_repositories_grpc(api_client)


async def gen_list_repositories_ephemeral_grpc(
    executable_path: str,
    python_file: Optional[str],
    module_name: Optional[str],
    working_directory: Optional[str],
    attribute: Optional[str],
    package_name: Optional[str],
) -> "ListRepositoriesResponse":
    from dagster._grpc.client import ephemeral_grpc_api_client

    check.str_param(executable_path, "executable_path")
    check.opt_str_param(python_file, "python_file")
    check.opt_str_param(module_name, "module_name")
    check.opt_str_param(working_directory, "working_directory")
    check.opt_str_param(attribute, "attribute")
    check.opt_str_param(package_name, "package_name")

    with ephemeral_grpc_api_client(
        loadable_target_origin=LoadableTargetOrigin(
            executable_path=executable_path,
            module_name=module_name,
            python_file=python_file,
            working_directory=working_directory,
            attribute=attribute,
            package_name=package_name,
        )
    ) as api_client:
        return await gen_list_repositories_grpc(api_client)



from collections.abc import Iterator, Mapping
from contextlib import ExitStack
from typing import AbstractSet, Any, Callable, Final, Optional, cast  # noqa: UP035

import sqlalchemy as db
import sqlalchemy.exc as db_exc
from sqlalchemy.engine import Connection
from tqdm import tqdm
from typing_extensions import TypeAlias

import dagster._check as check
from dagster._core.execution.backfill import (
    BULK_ACTION_TERMINAL_STATUSES,
    BulkActionStatus,
    PartitionBackfill,
)
from dagster._core.storage.dagster_run import DagsterRun, DagsterRunStatus, RunRecord, RunsFilter
from dagster._core.storage.runs.base import RunStorage
from dagster._core.storage.runs.schema import (
    BackfillTagsTable,
    BulkActionsTable,
    RunsTable,
    RunTagsTable,
)
from dagster._core.storage.sqlalchemy_compat import db_select
from dagster._core.storage.tags import (
    BACKFILL_ID_TAG,
    PARTITION_NAME_TAG,
    PARTITION_SET_TAG,
    REPOSITORY_LABEL_TAG,
)
from dagster._serdes import deserialize_value

RUN_PARTITIONS = "run_partitions"
RUN_START_END = (  # was run_start_end, but renamed to overwrite bad timestamps written
    "run_start_end_overwritten"
)
RUN_REPO_LABEL_TAGS = "run_repo_label_tags"
BULK_ACTION_TYPES = "bulk_action_types"
RUN_BACKFILL_ID = "run_backfill_id"
BACKFILL_JOB_NAME_AND_TAGS = "backfill_job_name_and_tags"
BACKFILL_END_TIMESTAMP = "backfill_end_timestamp"

PrintFn: TypeAlias = Callable[[Any], None]
MigrationFn: TypeAlias = Callable[[RunStorage, Optional[PrintFn]], None]

# for `dagster instance migrate`, paired with schema changes
REQUIRED_DATA_MIGRATIONS: Final[Mapping[str, Callable[[], MigrationFn]]] = {
    RUN_PARTITIONS: lambda: migrate_run_partition,
    RUN_REPO_LABEL_TAGS: lambda: migrate_run_repo_tags,
    BULK_ACTION_TYPES: lambda: migrate_bulk_actions,
    RUN_BACKFILL_ID: lambda: migrate_run_backfill_id,
    BACKFILL_JOB_NAME_AND_TAGS: lambda: migrate_backfill_job_name_and_tags,
    BACKFILL_END_TIMESTAMP: lambda: migrate_backfill_end_timestamp,
}
# for `dagster instance reindex`, optionally run for better read performance
OPTIONAL_DATA_MIGRATIONS: Final[Mapping[str, Callable[[], MigrationFn]]] = {
    RUN_START_END: lambda: migrate_run_start_end,
}

CHUNK_SIZE = 100

UNSTARTED_RUN_STATUSES: Final[AbstractSet[DagsterRunStatus]] = {
    DagsterRunStatus.QUEUED,
    DagsterRunStatus.NOT_STARTED,
    DagsterRunStatus.MANAGED,
    DagsterRunStatus.STARTING,
}


def chunked_run_iterator(
    storage: RunStorage, print_fn: Optional[PrintFn] = None, chunk_size: int = CHUNK_SIZE
) -> Iterator[DagsterRun]:
    with ExitStack() as stack:
        if print_fn:
            run_count = storage.get_runs_count()
            progress = stack.enter_context(tqdm(total=run_count))
        else:
            progress = None

        cursor = None
        has_more = True

        while has_more:
            chunk = storage.get_runs(cursor=cursor, limit=chunk_size)
            has_more = chunk_size and len(chunk) >= chunk_size

            for run in chunk:
                cursor = run.run_id
                yield run

            if progress:
                progress.update(len(chunk))


def chunked_run_records_iterator(
    storage: RunStorage, print_fn: Optional[PrintFn] = None, chunk_size: int = CHUNK_SIZE
) -> Iterator[RunRecord]:
    with ExitStack() as stack:
        if print_fn:
            run_count = storage.get_runs_count()
            progress = stack.enter_context(tqdm(total=run_count))
        else:
            progress = None

        cursor = None
        has_more = True

        while has_more:
            chunk = storage.get_run_records(cursor=cursor, limit=chunk_size)
            has_more = chunk_size and len(chunk) >= chunk_size

            for run in chunk:
                cursor = run.dagster_run.run_id
                yield run

            if progress:
                progress.update(len(chunk))


def chunked_backfill_iterator(
    storage: RunStorage, print_fn: Optional[PrintFn] = None, chunk_size: int = CHUNK_SIZE
) -> Iterator[PartitionBackfill]:
    with ExitStack() as stack:
        if print_fn:
            backfill_count = storage.get_backfills_count()
            progress = stack.enter_context(tqdm(total=backfill_count))
        else:
            progress = None

        cursor = None
        has_more = True

        while has_more:
            chunk = storage.get_backfills(cursor=cursor, limit=chunk_size)
            has_more = chunk_size and len(chunk) >= chunk_size

            for backfill in chunk:
                cursor = backfill.backfill_id
                yield backfill

            if progress:
                progress.update(len(chunk))


def migrate_run_partition(storage: RunStorage, print_fn: Optional[PrintFn] = None) -> None:
    """Utility method to build an asset key index from the data in existing event log records.
    Takes in event_log_storage, and a print_fn to keep track of progress.
    """
    if print_fn:
        print_fn("Querying run storage.")

    for run in chunked_run_iterator(storage, print_fn):
        if PARTITION_NAME_TAG not in run.tags:
            continue
        if PARTITION_SET_TAG not in run.tags:
            continue

        storage.add_run_tags(run.run_id, run.tags)


def migrate_run_start_end(storage: RunStorage, print_fn: Optional[PrintFn] = None) -> None:
    """Utility method that updates the start and end times of historical runs using the completed event log."""
    if print_fn:
        print_fn("Querying run and event log storage.")

    for run_record in chunked_run_records_iterator(storage, print_fn):
        if run_record.dagster_run.status in UNSTARTED_RUN_STATUSES:
            continue

        # commented out here to ensure that previously written timestamps that may not have
        # standardized to UTC would get overwritten
        # if run_record.start_time:
        #     continue

        add_run_stats(storage, run_record.dagster_run.run_id)


def add_run_stats(run_storage: RunStorage, run_id: str) -> None:
    from dagster._core.instance import DagsterInstance
    from dagster._core.storage.runs.sql_run_storage import SqlRunStorage

    check.str_param(run_id, "run_id")
    check.inst_param(run_storage, "run_storage", RunStorage)

    if not isinstance(run_storage, SqlRunStorage):
        return

    instance = check.inst_param(run_storage._instance, "instance", DagsterInstance)  # noqa: SLF001
    run_stats = instance.get_run_stats(run_id)

    with run_storage.connect() as conn:
        conn.execute(
            RunsTable.update()
            .where(RunsTable.c.run_id == run_id)
            .values(
                start_time=run_stats.start_time,
                end_time=run_stats.end_time,
            )
        )


def migrate_run_repo_tags(run_storage: RunStorage, print_fn: Optional[PrintFn] = None) -> None:
    from dagster._core.storage.runs.sql_run_storage import SqlRunStorage

    if not isinstance(run_storage, SqlRunStorage):
        return

    if print_fn:
        print_fn("Querying run storage.")

    subquery = (
        db_select([RunTagsTable.c.run_id.label("tags_run_id")])
        .where(RunTagsTable.c.key == REPOSITORY_LABEL_TAG)
        .alias("tag_subquery")
    )
    base_query = (
        db_select([RunsTable.c.run_body, RunsTable.c.id])
        .select_from(
            RunsTable.join(subquery, RunsTable.c.run_id == subquery.c.tags_run_id, isouter=True)
        )
        .where(subquery.c.tags_run_id.is_(None))
        .order_by(db.asc(RunsTable.c.id))
        .limit(CHUNK_SIZE)
    )

    cursor = None
    has_more = True
    while has_more:
        if cursor:
            query = base_query.where(RunsTable.c.id > cursor)
        else:
            query = base_query

        with run_storage.connect() as conn:
            result_proxy = conn.execute(query)
            rows = result_proxy.fetchall()
            result_proxy.close()

            has_more = len(rows) >= CHUNK_SIZE
            for row in rows:
                run = deserialize_value(cast(str, row[0]), DagsterRun)
                cursor = row[1]
                write_repo_tag(conn, run)


def write_repo_tag(conn: Connection, run: DagsterRun) -> None:
    if not run.remote_job_origin:
        # nothing to do
        return

    repository_label = run.remote_job_origin.repository_origin.get_label()
    try:
        conn.execute(
            RunTagsTable.insert().values(
                run_id=run.run_id,
                key=REPOSITORY_LABEL_TAG,
                value=repository_label,
            )
        )
    except db_exc.IntegrityError:
        # tag already exists, swallow
        pass


def migrate_bulk_actions(run_storage: RunStorage, print_fn: Optional[PrintFn] = None) -> None:
    from dagster._core.storage.runs.sql_run_storage import SqlRunStorage

    if not isinstance(run_storage, SqlRunStorage):
        return

    if print_fn:
        print_fn("Querying run storage.")

    base_query = (
        db_select([BulkActionsTable.c.body, BulkActionsTable.c.id])
        .where(BulkActionsTable.c.action_type.is_(None))
        .order_by(db.asc(BulkActionsTable.c.id))
        .limit(CHUNK_SIZE)
    )

    cursor = None
    has_more = True
    while has_more:
        if cursor:
            query = base_query.where(BulkActionsTable.c.id > cursor)
        else:
            query = base_query

        with run_storage.connect() as conn:
            result_proxy = conn.execute(query)
            rows = result_proxy.fetchall()
            result_proxy.close()

            has_more = len(rows) >= CHUNK_SIZE
            for row in rows:
                backfill = deserialize_value(row[0], PartitionBackfill)  # type: ignore  # (pyright bug)
                storage_id = row[1]
                conn.execute(
                    BulkActionsTable.update()
                    .values(
                        selector_id=backfill.selector_id,
                        action_type=backfill.bulk_action_type.value,
                    )
                    .where(BulkActionsTable.c.id == storage_id)
                )
                cursor = storage_id


def migrate_run_backfill_id(storage: RunStorage, print_fn: Optional[PrintFn] = None) -> None:
    """Utility method to add a backfill_id column to the runs table and populate it with the backfill_id of the run."""
    from dagster._core.storage.runs.sql_run_storage import SqlRunStorage

    if print_fn:
        print_fn("Querying run storage.")

    check.inst_param(storage, "run_storage", RunStorage)

    if not isinstance(storage, SqlRunStorage):
        return

    with storage.connect() as conn:
        cursor = None
        has_more = True
        base_query = (
            db_select([RunTagsTable.c.id, RunTagsTable.c.run_id, RunTagsTable.c.value])
            .where(RunTagsTable.c.key == BACKFILL_ID_TAG)
            .limit(CHUNK_SIZE)
            .order_by(RunTagsTable.c.id)
        )

        while has_more:
            if cursor:
                query = base_query.where(RunTagsTable.c.id > cursor)
            else:
                query = base_query
            res = storage.fetchall(query)
            has_more = len(res) >= CHUNK_SIZE
            for row in res:
                cursor = row["id"]
                run_id = row["run_id"]
                backfill_id = row["value"]

                add_backfill_id(conn, run_id=run_id, backfill_id=backfill_id)


def add_backfill_id(conn, run_id: str, backfill_id: str) -> None:
    check.str_param(run_id, "run_id")
    check.str_param(backfill_id, "backfill_id")
    conn.execute(
        RunsTable.update()
        .where(RunsTable.c.run_id == run_id)
        .where(RunsTable.c.backfill_id.is_(None))
        .values(
            backfill_id=backfill_id,
        )
    )


def migrate_backfill_job_name_and_tags(
    storage: RunStorage, print_fn: Optional[PrintFn] = None
) -> None:
    """Utility method to add a backfill's job_name to the bulk_actions table and tags to the backfill_tags table."""
    if print_fn:
        print_fn("Querying run storage.")

    for backfill in chunked_backfill_iterator(storage, print_fn):
        if backfill.tags:
            add_backfill_tags(
                run_storage=storage, backfill_id=backfill.backfill_id, tags=backfill.tags
            )

        if backfill.job_name is not None:
            add_backfill_job_name(
                run_storage=storage, backfill_id=backfill.backfill_id, job_name=backfill.job_name
            )


def add_backfill_tags(run_storage: RunStorage, backfill_id: str, tags: Mapping[str, str]):
    from dagster._core.storage.runs.sql_run_storage import SqlRunStorage

    check.str_param(backfill_id, "run_id")
    check.dict_param(tags, "tags", key_type=str, value_type=str)
    check.inst_param(run_storage, "run_storage", RunStorage)

    if not isinstance(run_storage, SqlRunStorage):
        return

    with run_storage.connect() as conn:
        conn.execute(
            BackfillTagsTable.insert(),
            [
                dict(
                    backfill_id=backfill_id,
                    key=k,
                    value=v,
                )
                for k, v in tags.items()
            ],
        )


def add_backfill_job_name(run_storage: RunStorage, backfill_id: str, job_name: str):
    from dagster._core.storage.runs.sql_run_storage import SqlRunStorage

    check.str_param(backfill_id, "run_id")
    check.str_param(job_name, "job_name")
    check.inst_param(run_storage, "run_storage", RunStorage)

    if not isinstance(run_storage, SqlRunStorage):
        return

    with run_storage.connect() as conn:
        conn.execute(
            BulkActionsTable.update()
            .values(
                job_name=job_name,
            )
            .where(BulkActionsTable.c.key == backfill_id)
        )


def migrate_backfill_end_timestamp(storage: RunStorage, print_fn: Optional[PrintFn] = None) -> None:
    """Utility method to add a backfill's end timestamp to the serialized backfill stored in the bulk actions table."""
    if print_fn:
        print_fn("Querying run storage.")

    for backfill in chunked_backfill_iterator(storage, print_fn):
        if backfill.status == BulkActionStatus.REQUESTED:
            # we don't want to mutate a backfill that is still in progress. Additionally, it won't
            # have an end timestamp until it moves to a terminal state
            continue
        if backfill.backfill_end_timestamp is None:
            end_time = get_end_timestamp_for_backfill(storage, backfill)
            updated_backfill = backfill.with_end_timestamp(end_time)
            storage.update_backfill(updated_backfill)


def get_end_timestamp_for_backfill(run_storage: RunStorage, backfill: PartitionBackfill) -> float:
    assert backfill.status in BULK_ACTION_TERMINAL_STATUSES
    filters = RunsFilter.for_backfill(backfill.backfill_id)
    run_records = run_storage.get_run_records(filters=filters)
    if len(run_records) == 0:
        # backfill was moved to a terminal state before any runs were launched. We cannot
        # reconstruct the time the backfill actually moved to a terminal state, so use the start
        # time as an estimation
        return backfill.backfill_timestamp
    return max([record.end_time or 0 for record in run_records])



from typing import TypeVar, cast

from typing_extensions import NotRequired, get_origin, is_typeddict

from dagster._utils.typing_api import is_closed_python_optional_type

_TypedDictClass = TypeVar("_TypedDictClass")


def init_optional_typeddict(cls: type[_TypedDictClass]) -> _TypedDictClass:
    """Initialize a TypedDict with optional values."""
    if not is_typeddict(cls):
        raise Exception("Must pass a TypedDict class to init_optional_typeddict")
    result = {}
    for key, value in cls.__annotations__.items():
        # If the value is a typed dict, recursively initialize it
        if is_typeddict(value):
            result[key] = init_optional_typeddict(value)
        elif is_closed_python_optional_type(value):
            result[key] = None
        elif get_origin(value) is dict:
            result[key] = {}
        elif get_origin(value) is NotRequired:
            continue
        else:
            raise Exception("fields must be either optional or typed dicts")
    return cast(_TypedDictClass, result)



"""add columns action_type and selector_id to bulk_actions.

Revision ID: 6860f830e40c
Revises: 721d858e1dda
Create Date: 2022-05-20 15:00:01.260860

"""

import sqlalchemy as db
from alembic import op
from dagster._core.storage.migration.utils import has_column, has_index, has_table

# revision identifiers, used by Alembic.
revision = "6860f830e40c"
down_revision = "721d858e1dda"
branch_labels = None
depends_on = None


def upgrade():
    if not has_table("bulk_actions"):
        return

    if not has_column("bulk_actions", "action_type"):
        op.add_column("bulk_actions", db.Column("action_type", db.String(32), nullable=True))

    if not has_column("bulk_actions", "selector_id"):
        op.add_column("bulk_actions", db.Column("selector_id", db.Text, nullable=True))

    if not has_index("bulk_actions", "idx_bulk_actions_action_type"):
        op.create_index(
            "idx_bulk_actions_action_type",
            "bulk_actions",
            ["action_type"],
            unique=False,
            mysql_length={"action_type": 32},
        )

    if not has_index("bulk_actions", "idx_bulk_actions_selector_id"):
        op.create_index(
            "idx_bulk_actions_selector_id",
            "bulk_actions",
            ["selector_id"],
            unique=False,
            mysql_length={"selector_id": 64},
        )


def downgrade():
    with op.batch_alter_table("bulk_actions") as batch_op:
        if has_index("bulk_actions", "idx_bulk_actions_action_type"):
            batch_op.drop_index("idx_bulk_actions_action_type")
        if has_index("bulk_actions", "idx_bulk_actions_selector_id"):
            batch_op.drop_index("idx_bulk_actions_selector_id")
        if has_column("bulk_actions", "action_type"):
            batch_op.drop_column("action_type")
        if has_column("bulk_actions", "selector_id"):
            batch_op.drop_column("selector_id")



import copy
from collections.abc import Mapping, Sequence
from functools import update_wrapper
from typing import TYPE_CHECKING, Callable, Optional, Union, cast

import dagster._check as check
from dagster._core.definitions.metadata import RawMetadataMapping
from dagster._core.definitions.resource_annotation import get_resource_args
from dagster._core.definitions.run_request import RunRequest, SkipReason
from dagster._core.definitions.schedule_definition import (
    DecoratedScheduleFunction,
    DefaultScheduleStatus,
    RawScheduleEvaluationFunction,
    RunRequestIterator,
    ScheduleDefinition,
    ScheduleEvaluationContext,
    has_at_least_one_parameter,
    validate_and_get_schedule_resource_dict,
)
from dagster._core.definitions.sensor_definition import get_context_param_name
from dagster._core.definitions.target import ExecutableDefinition
from dagster._core.errors import (
    DagsterInvalidDefinitionError,
    ScheduleExecutionError,
    user_code_error_boundary,
)
from dagster._utils import ensure_gen
from dagster._utils.tags import normalize_tags

if TYPE_CHECKING:
    from dagster._core.definitions.asset_selection import CoercibleToAssetSelection
    from dagster._core.definitions.assets import AssetsDefinition
    from dagster._core.definitions.job_definition import JobDefinition
    from dagster._core.definitions.unresolved_asset_job_definition import (
        UnresolvedAssetJobDefinition,
    )


def schedule(
    cron_schedule: Union[str, Sequence[str]],
    *,
    job_name: Optional[str] = None,
    name: Optional[str] = None,
    tags: Optional[Mapping[str, str]] = None,
    tags_fn: Optional[Callable[[ScheduleEvaluationContext], Optional[Mapping[str, str]]]] = None,
    metadata: Optional[RawMetadataMapping] = None,
    should_execute: Optional[Callable[[ScheduleEvaluationContext], bool]] = None,
    environment_vars: Optional[Mapping[str, str]] = None,
    execution_timezone: Optional[str] = None,
    description: Optional[str] = None,
    job: Optional[ExecutableDefinition] = None,
    default_status: DefaultScheduleStatus = DefaultScheduleStatus.STOPPED,
    required_resource_keys: Optional[set[str]] = None,
    target: Optional[
        Union[
            "CoercibleToAssetSelection",
            "AssetsDefinition",
            "JobDefinition",
            "UnresolvedAssetJobDefinition",
        ]
    ] = None,
) -> Callable[[RawScheduleEvaluationFunction], ScheduleDefinition]:
    """Creates a schedule following the provided cron schedule and requests runs for the provided job.

    The decorated function takes in a :py:class:`~dagster.ScheduleEvaluationContext` as its only
    argument, and does one of the following:

    1. Return a :py:class:`~dagster.RunRequest` object.
    2. Return a list of :py:class:`~dagster.RunRequest` objects.
    3. Return a :py:class:`~dagster.SkipReason` object, providing a descriptive message of why no runs were requested.
    4. Return nothing (skipping without providing a reason)
    5. Return a run config dictionary.
    6. Yield a :py:class:`~dagster.SkipReason` or yield one ore more :py:class:`~dagster.RunRequest` objects.

    Returns a :py:class:`~dagster.ScheduleDefinition`.

    Args:
        cron_schedule (Union[str, Sequence[str]]): A valid cron string or sequence of cron strings
            specifying when the schedule will run, e.g., ``45 23 * * 6`` for a schedule that runs
            at 11:45 PM every Saturday. If a sequence is provided, then the schedule will run for
            the union of all execution times for the provided cron strings, e.g.,
            ``['45 23 * * 6', '30 9 * * 0']`` for a schedule that runs at 11:45 PM every Saturday and
            9:30 AM every Sunday.
        name (Optional[str]): The name of the schedule.
        tags (Optional[Mapping[str, str]]): A set of key-value tags that annotate the schedule and can
            be used for searching and filtering in the UI.
        tags_fn (Optional[Callable[[ScheduleEvaluationContext], Optional[Dict[str, str]]]]): A function
            that generates tags to attach to the schedule's runs. Takes a
            :py:class:`~dagster.ScheduleEvaluationContext` and returns a dictionary of tags (string
            key-value pairs). **Note**: Either ``tags`` or ``tags_fn`` may be set, but not both.
        metadata (Optional[Mapping[str, Any]]): A set of metadata entries that annotate the
            schedule. Values will be normalized to typed `MetadataValue` objects.
        should_execute (Optional[Callable[[ScheduleEvaluationContext], bool]]): A function that runs at
            schedule execution time to determine whether a schedule should execute or skip. Takes a
            :py:class:`~dagster.ScheduleEvaluationContext` and returns a boolean (``True`` if the
            schedule should execute). Defaults to a function that always returns ``True``.
        execution_timezone (Optional[str]): Timezone in which the schedule should run.
            Supported strings for timezones are the ones provided by the
            `IANA time zone database <https://www.iana.org/time-zones>`_ - e.g. ``"America/Los_Angeles"``.
        description (Optional[str]): A human-readable description of the schedule.
        job (Optional[Union[GraphDefinition, JobDefinition, UnresolvedAssetJobDefinition]]): The job
            that should execute when the schedule runs.
        default_status (DefaultScheduleStatus): If set to ``RUNNING``, the schedule will immediately be active when starting Dagster. The default status can be overridden from the Dagster UI or via the GraphQL API.
        required_resource_keys (Optional[Set[str]]): The set of resource keys required by the schedule.
        target (Optional[Union[CoercibleToAssetSelection, AssetsDefinition, JobDefinition, UnresolvedAssetJobDefinition]]):
            The target that the schedule will execute.
            It can take :py:class:`~dagster.AssetSelection` objects and anything coercible to it (e.g. `str`, `Sequence[str]`, `AssetKey`, `AssetsDefinition`).
            It can also accept :py:class:`~dagster.JobDefinition` (a function decorated with `@job` is an instance of `JobDefinition`) and `UnresolvedAssetJobDefinition` (the return value of :py:func:`~dagster.define_asset_job`) objects.
            This parameter will replace `job` and `job_name`.
    """

    def inner(fn: RawScheduleEvaluationFunction) -> ScheduleDefinition:
        from dagster._config.pythonic_config import validate_resource_annotated_function

        check.callable_param(fn, "fn")
        validate_resource_annotated_function(fn)

        schedule_name = name or fn.__name__

        validated_tags = None

        # perform upfront validation of schedule tags
        if tags_fn and tags:
            raise DagsterInvalidDefinitionError(
                "Attempted to provide both tags_fn and tags as arguments"
                " to ScheduleDefinition. Must provide only one of the two."
            )
        elif tags:
            validated_tags = normalize_tags(
                tags, allow_private_system_tags=False, warning_stacklevel=3
            )

        context_param_name = get_context_param_name(fn)
        resource_arg_names: set[str] = {arg.name for arg in get_resource_args(fn)}

        def _wrapped_fn(context: ScheduleEvaluationContext) -> RunRequestIterator:
            if should_execute:
                with user_code_error_boundary(
                    ScheduleExecutionError,
                    lambda: (
                        "Error occurred during the execution of should_execute for schedule"
                        f" {schedule_name}"
                    ),
                ):
                    if not should_execute(context):
                        yield SkipReason(
                            f"should_execute function for {schedule_name} returned false."
                        )
                        return
            resources = validate_and_get_schedule_resource_dict(
                context.resources, schedule_name, resource_arg_names
            )

            with user_code_error_boundary(
                ScheduleExecutionError,
                lambda: f"Error occurred during the evaluation of schedule {schedule_name}",
            ):
                context_param = {context_param_name: context} if context_param_name else {}
                result = fn(**context_param, **resources)

                if isinstance(result, dict):
                    # this is the run-config based decorated function, wrap the evaluated run config
                    # and tags in a RunRequest
                    evaluated_run_config = copy.deepcopy(result)
                    evaluated_tags = (
                        validated_tags
                        or (
                            tags_fn
                            and normalize_tags(tags_fn(context), allow_private_system_tags=False)
                        )
                        or None
                    )
                    yield RunRequest(
                        run_key=None,
                        run_config=evaluated_run_config,
                        tags=evaluated_tags,
                    )
                elif isinstance(result, list):
                    yield from cast(list[RunRequest], result)
                else:
                    # this is a run-request based decorated function
                    yield from cast(RunRequestIterator, ensure_gen(result))

        has_context_arg = has_at_least_one_parameter(fn)
        evaluation_fn = DecoratedScheduleFunction(
            decorated_fn=fn,
            wrapped_fn=_wrapped_fn,
            has_context_arg=has_context_arg,
        )

        schedule_def = ScheduleDefinition.dagster_internal_init(
            name=schedule_name,
            cron_schedule=cron_schedule,
            job_name=job_name,
            environment_vars=environment_vars,
            execution_timezone=execution_timezone,
            description=description,
            execution_fn=evaluation_fn,
            job=job,
            default_status=default_status,
            required_resource_keys=required_resource_keys,
            run_config=None,  # cannot supply run_config or run_config_fn to decorator
            run_config_fn=None,
            tags=tags,
            tags_fn=None,  # cannot supply tags or tags_fn to decorator
            metadata=metadata,
            should_execute=None,  # already encompassed in evaluation_fn
            target=target,
        )

        update_wrapper(schedule_def, wrapped=fn)

        return schedule_def

    return inner



from dagster._config.config_schema import (
    ConfigSchema as ConfigSchema,
    UserConfigSchema as UserConfigSchema,
)
from dagster._config.config_type import (
    ALL_CONFIG_BUILTINS as ALL_CONFIG_BUILTINS,
    Array as Array,
    Bool as Bool,
    ConfigAnyInstance as ConfigAnyInstance,
    ConfigBoolInstance as ConfigBoolInstance,
    ConfigFloatInstance as ConfigFloatInstance,
    ConfigIntInstance as ConfigIntInstance,
    ConfigScalar as ConfigScalar,
    ConfigScalarKind as ConfigScalarKind,
    ConfigStringInstance as ConfigStringInstance,
    ConfigType as ConfigType,
    ConfigTypeKind as ConfigTypeKind,
    Enum as Enum,
    EnumValue as EnumValue,
    Float as Float,
    Int as Int,
    Noneable as Noneable,
    ScalarUnion as ScalarUnion,
    String as String,
    get_builtin_scalar_by_name as get_builtin_scalar_by_name,
)

# Separate section necessary to prevent circular import
from dagster._config.errors import (
    DagsterEvaluationErrorReason as DagsterEvaluationErrorReason,
    EvaluationError as EvaluationError,
    FieldNotDefinedErrorData as FieldNotDefinedErrorData,
    FieldsNotDefinedErrorData as FieldsNotDefinedErrorData,
    MissingFieldErrorData as MissingFieldErrorData,
    MissingFieldsErrorData as MissingFieldsErrorData,
    PostProcessingError as PostProcessingError,
    RuntimeMismatchErrorData as RuntimeMismatchErrorData,
    SelectorTypeErrorData as SelectorTypeErrorData,
)
from dagster._config.evaluate_value_result import EvaluateValueResult as EvaluateValueResult
from dagster._config.field import (
    Field as Field,
    resolve_to_config_type as resolve_to_config_type,
)
from dagster._config.field_utils import (
    FIELD_NO_DEFAULT_PROVIDED as FIELD_NO_DEFAULT_PROVIDED,
    Map as Map,
    Permissive as Permissive,
    Selector as Selector,
    Shape as Shape,
    compute_fields_hash as compute_fields_hash,
    convert_potential_field as convert_potential_field,
)
from dagster._config.post_process import (
    post_process_config as post_process_config,
    resolve_defaults as resolve_defaults,
)
from dagster._config.primitive_mapping import (
    is_supported_config_python_builtin as is_supported_config_python_builtin,
)
from dagster._config.snap import (
    ConfigEnumValueSnap as ConfigEnumValueSnap,
    ConfigFieldSnap as ConfigFieldSnap,
    ConfigSchemaSnapshot as ConfigSchemaSnapshot,
    ConfigTypeSnap as ConfigTypeSnap,
    get_recursive_type_keys as get_recursive_type_keys,
    snap_from_config_type as snap_from_config_type,
    snap_from_field as snap_from_field,
)
from dagster._config.source import (
    BoolSource as BoolSource,
    BoolSourceType as BoolSourceType,
    IntSource as IntSource,
    IntSourceType as IntSourceType,
    StringSource as StringSource,
    StringSourceType as StringSourceType,
)
from dagster._config.stack import (
    EvaluationStackListItemEntry as EvaluationStackListItemEntry,
    EvaluationStackMapKeyEntry as EvaluationStackMapKeyEntry,
    EvaluationStackMapValueEntry as EvaluationStackMapValueEntry,
    EvaluationStackPathEntry as EvaluationStackPathEntry,
)
from dagster._config.type_printer import print_config_type_to_string as print_config_type_to_string
from dagster._config.validate import (
    process_config as process_config,
    validate_config as validate_config,
    validate_config_from_snap as validate_config_from_snap,
)



import itertools
from collections import defaultdict
from collections.abc import Mapping
from functools import cached_property
from typing import AbstractSet, Optional, cast  # noqa: UP035

import dagster._check as check
from dagster._core.definitions.asset_key import AssetCheckKey, AssetKey, EntityKey
from dagster._core.definitions.asset_spec import AssetExecutionType
from dagster._core.definitions.backfill_policy import BackfillPolicy
from dagster._core.definitions.dependency import NodeHandle, NodeOutputHandle
from dagster._core.definitions.graph_definition import GraphDefinition, SubselectedGraphDefinition
from dagster._core.definitions.node_definition import NodeDefinition
from dagster._core.definitions.op_selection import get_graph_subset
from dagster._core.errors import DagsterInvalidInvocationError
from dagster._core.utils import toposort_flatten
from dagster._record import IHaveNew, copy, record, record_custom


@record_custom
class AssetGraphComputation(IHaveNew):
    """A computation whose purpose is to materialize assets, observe assets, and/or evaluate asset
    checks.

    Binds a NodeDefinition to the asset keys and asset check keys that it interacts with.
    """

    node_def: NodeDefinition
    keys_by_input_name: Mapping[str, AssetKey]
    keys_by_output_name: Mapping[str, AssetKey]
    check_keys_by_output_name: Mapping[str, AssetCheckKey]
    backfill_policy: Optional[BackfillPolicy]
    can_subset: bool
    is_subset: bool
    selected_asset_keys: AbstractSet[AssetKey]
    selected_asset_check_keys: AbstractSet[AssetCheckKey]
    execution_type: AssetExecutionType

    def __new__(
        cls,
        node_def: NodeDefinition,
        keys_by_input_name: Mapping[str, AssetKey],
        keys_by_output_name: Mapping[str, AssetKey],
        check_keys_by_output_name: Mapping[str, AssetCheckKey],
        backfill_policy: Optional[BackfillPolicy],
        can_subset: bool,
        is_subset: bool,
        selected_asset_keys: AbstractSet[AssetKey],
        selected_asset_check_keys: AbstractSet[AssetCheckKey],
        execution_type: AssetExecutionType,
    ):
        result = super().__new__(
            cls,
            node_def=node_def,
            keys_by_input_name=keys_by_input_name,
            keys_by_output_name=keys_by_output_name,
            check_keys_by_output_name=check_keys_by_output_name,
            backfill_policy=backfill_policy,
            can_subset=can_subset,
            is_subset=is_subset,
            selected_asset_keys=selected_asset_keys,
            selected_asset_check_keys=selected_asset_check_keys,
            execution_type=execution_type,
        )
        result.output_names_by_key  # eagerly compute this to catch any errors at contruction time  # noqa: B018
        return result

    @cached_property
    def output_names_by_key(self) -> Mapping[AssetKey, str]:
        output_names_by_key: dict[AssetKey, str] = {}
        for output_name, key in self.keys_by_output_name.items():
            if key in output_names_by_key:
                check.failed(
                    f"Outputs '{output_names_by_key[key]}' and '{output_name}' both target the "
                    "same asset key. Each asset key should correspond to a single output."
                )
            output_names_by_key[key] = output_name

        return output_names_by_key

    @cached_property
    def entity_keys_by_op_output_handle(
        self,
    ) -> Mapping[NodeOutputHandle, EntityKey]:
        result = {}
        for output_name, key in itertools.chain(
            self.keys_by_output_name.items(), self.check_keys_by_output_name.items()
        ):
            output_def, node_handle = self.full_node_def.resolve_output_to_origin(output_name, None)
            result[NodeOutputHandle(node_handle=node_handle, output_name=output_def.name)] = key

        return result

    @property
    def full_node_def(self) -> NodeDefinition:
        node_def = self.node_def
        while isinstance(node_def, SubselectedGraphDefinition):
            node_def = node_def.parent_graph_def

        return node_def

    def subset_for(
        self,
        selected_asset_keys: AbstractSet[AssetKey],
        selected_asset_check_keys: Optional[AbstractSet[AssetCheckKey]],
    ) -> "AssetGraphComputation":
        check.invariant(
            self.can_subset,
            f"Attempted to subset AssetsDefinition for {self.node_def.name}, but can_subset=False.",
        )

        # Set of assets within selected_asset_keys which are outputted by this AssetDefinition
        asset_subselection = selected_asset_keys & self.selected_asset_keys
        if selected_asset_check_keys is None:
            # filter to checks that target selected asset keys
            asset_check_subselection = {
                key for key in self.selected_asset_check_keys if key.asset_key in asset_subselection
            }
        else:
            asset_check_subselection = selected_asset_check_keys & self.selected_asset_check_keys

        # Early escape if all assets and checks in AssetsDefinition are selected
        if (
            asset_subselection == self.selected_asset_keys
            and asset_check_subselection == self.selected_asset_check_keys
        ):
            return self
        elif isinstance(self.node_def, GraphDefinition):  # Node is graph-backed asset
            subsetted_node = self._subset_graph_backed_asset(
                asset_subselection, asset_check_subselection
            )

            return copy(
                self,
                node_def=subsetted_node,
                selected_asset_keys=selected_asset_keys & self.selected_asset_keys,
                selected_asset_check_keys=asset_check_subselection,
                is_subset=True,
            )
        else:
            # multi_asset subsetting
            return copy(
                self,
                selected_asset_keys=asset_subselection,
                selected_asset_check_keys=asset_check_subselection,
                is_subset=True,
            )

    def _subset_graph_backed_asset(
        self,
        selected_asset_keys: AbstractSet[AssetKey],
        selected_asset_check_keys: AbstractSet[AssetCheckKey],
    ) -> SubselectedGraphDefinition:
        if not isinstance(self.node_def, GraphDefinition):
            raise DagsterInvalidInvocationError(
                "Method _subset_graph_backed_asset cannot subset an asset that is not a graph"
            )

        dep_op_handles_by_entity_key = cast(
            # because self.node_def is a graph, these NodeHandles that reference ops inside it will
            # not be None
            Mapping[EntityKey, AbstractSet[NodeHandle]],
            self.dep_op_handles_by_entity_key,
        )
        op_selection: list[str] = []
        for asset_key in selected_asset_keys:
            dep_node_handles = dep_op_handles_by_entity_key[asset_key]
            for dep_op_handle in dep_node_handles:
                op_selection.append(".".join(dep_op_handle.path))
        for asset_check_key in selected_asset_check_keys:
            dep_op_handles = dep_op_handles_by_entity_key[asset_check_key]
            for dep_op_handle in dep_op_handles:
                op_selection.append(".".join(dep_op_handle.path))

        selected_outputs_by_op_handle: dict[NodeHandle, set[str]] = defaultdict(set)
        for (
            op_output_handle,
            entity_keys,
        ) in self.entity_keys_by_dep_op_output_handle.items():
            if any(
                key in selected_asset_keys or key in selected_asset_check_keys
                for key in entity_keys
            ):
                selected_outputs_by_op_handle[op_output_handle.node_handle].add(
                    op_output_handle.output_name
                )

        return get_graph_subset(
            self.node_def, op_selection, selected_outputs_by_op_handle=selected_outputs_by_op_handle
        )

    @cached_property
    def dep_op_handles_by_entity_key(
        self,
    ) -> Mapping[EntityKey, AbstractSet[Optional[NodeHandle]]]:
        result = defaultdict(set)
        for op_output_handle, keys in self.entity_keys_by_dep_op_output_handle.items():
            for key in keys:
                result[key].add(op_output_handle.node_handle)

        return result

    @cached_property
    def entity_keys_by_dep_op_output_handle(
        self,
    ) -> Mapping[NodeOutputHandle, AbstractSet[EntityKey]]:
        """Returns a mapping between op outputs and the assets and asset checks that, when selected,
        those op outputs need to be produced for.
        For non-graph-backed assets, this is essentially the same as node_keys_by_output_name.
        For graph-backed multi-assets, depending on the asset selection, we often only need to
        execute a subset of the ops and for those ops to only produce a subset of their outputs,
        in order to materialize/observe the asset.
        Op output X will be selected when asset (or asset check) A is selected iff there's at least
        one path in the graph from X to the op output corresponding to A, and no outputs in that
        path directly correspond to other assets.
        """
        entity_keys_by_op_output_handle = self.entity_keys_by_op_output_handle
        if not isinstance(self.full_node_def, GraphDefinition):
            return {
                key: {op_output_handle}
                for key, op_output_handle in entity_keys_by_op_output_handle.items()
            }

        op_output_graph = OpOutputHandleGraph.from_graph(self.full_node_def)
        reverse_toposorted_op_outputs_handles = [
            *reversed(toposort_flatten(op_output_graph.upstream)),
            *(op_output_graph.op_output_handles - op_output_graph.upstream.keys()),
        ]

        result: dict[NodeOutputHandle, set[EntityKey]] = defaultdict(set)

        for op_output_handle in reverse_toposorted_op_outputs_handles:
            asset_key_or_check_key = entity_keys_by_op_output_handle.get(op_output_handle)

            if asset_key_or_check_key:
                result[op_output_handle].add(asset_key_or_check_key)
            else:
                child_op_output_handles = op_output_graph.downstream.get(op_output_handle, set())
                for child_op_output_handle in child_op_output_handles:
                    result[op_output_handle].update(result[child_op_output_handle])

        return result


@record
class OpOutputHandleGraph:
    """A graph where each node is a NodeOutputHandle corresponding to an op. There's an edge from
    op_output_1 to op_output_2 if op_output_2 is part of an op that has an input that's connected to
    op_output_1.
    """

    op_output_handles: AbstractSet[NodeOutputHandle]
    upstream: Mapping[NodeOutputHandle, AbstractSet[NodeOutputHandle]]
    downstream: Mapping[NodeOutputHandle, AbstractSet[NodeOutputHandle]]

    @staticmethod
    def from_graph(graph_def: "GraphDefinition") -> "OpOutputHandleGraph":
        op_output_handles = graph_def.get_op_output_handles(None)
        input_output_pairs = graph_def.get_op_input_output_handle_pairs(None)

        op_output_handles_by_op_handle: dict[NodeHandle, set[NodeOutputHandle]] = defaultdict(set)
        for op_output_handle in op_output_handles:
            op_output_handles_by_op_handle[op_output_handle.node_handle].add(op_output_handle)

        downstream_op_output_handles_by_op_output_handle: dict[
            NodeOutputHandle, set[NodeOutputHandle]
        ] = defaultdict(set)
        upstream_op_output_handles_by_op_output_handle: dict[
            NodeOutputHandle, set[NodeOutputHandle]
        ] = defaultdict(set)

        for op_output_handle, op_input_handle in input_output_pairs:
            downstream_op_output_handles = op_output_handles_by_op_handle[
                op_input_handle.node_handle
            ]
            for downstream_op_output_handle in downstream_op_output_handles:
                upstream_op_output_handles_by_op_output_handle[downstream_op_output_handle].add(
                    op_output_handle
                )
                downstream_op_output_handles_by_op_output_handle[op_output_handle].add(
                    downstream_op_output_handle
                )

        return OpOutputHandleGraph(
            op_output_handles=op_output_handles,
            downstream=downstream_op_output_handles_by_op_output_handle,
            upstream=upstream_op_output_handles_by_op_output_handle,
        )



from abc import ABC, abstractmethod
from collections.abc import Iterable, Mapping
from functools import cache
from typing import AbstractSet, Any, Optional  # noqa: UP035

from typing_extensions import TypeVar

from dagster import _check as check
from dagster._core.definitions.metadata.metadata_value import (
    MetadataValue,
    TableColumn as TableColumn,
    TableColumnConstraints as TableColumnConstraints,
    TableColumnLineage,
    TableSchema,
)
from dagster._model import DagsterModel
from dagster._model.pydantic_compat_layer import model_fields
from dagster._utils.typing_api import flatten_unions

# Python types that have a MetadataValue types that directly wraps them
DIRECTLY_WRAPPED_METADATA_TYPES = {
    str,
    float,
    int,
    bool,
    TableSchema,
    TableColumnLineage,
    type(None),
}

T_NamespacedKVSet = TypeVar("T_NamespacedKVSet", bound="NamespacedKVSet")


def is_raw_metadata_type(t: type) -> bool:
    return issubclass(t, MetadataValue) or t in DIRECTLY_WRAPPED_METADATA_TYPES


class NamespacedKVSet(ABC, DagsterModel):
    """Base class for defining a set of key-value pairs in the same namespace.

    Includes shared behavior between NamespacedMetadataSet and NamespacedTagSet.

    """

    @classmethod
    @abstractmethod
    def namespace(cls) -> str:
        raise NotImplementedError()

    @classmethod
    def _namespaced_key(cls, key: str) -> str:
        return f"{cls.namespace()}/{key}"

    @staticmethod
    def _strip_namespace_from_key(key: str) -> str:
        return key.split("/", 1)[1]

    def keys(self) -> Iterable[str]:
        return [
            self._namespaced_key(key)
            for key in model_fields(self).keys()
            # getattr returns the pydantic property on the subclass
            if getattr(self, key) is not None
        ]

    def __getitem__(self, key: str) -> Any:
        # getattr returns the pydantic property on the subclass
        return getattr(self, self._strip_namespace_from_key(key))

    @classmethod
    def extract(cls: type[T_NamespacedKVSet], values: Mapping[str, Any]) -> T_NamespacedKVSet:
        """Extracts entries from the provided dictionary into an instance of this class.

        Ignores any entries in the dictionary whose keys don't correspond to fields on this
        class.

        In general, the following should always pass:

        .. code-block:: python

            class MyKVSet(NamespacedKVSet):
                ...

            metadata: MyKVSet  = ...
            assert MyKVSet.extract(dict(metadata)) == metadata

        Args:
            values (Mapping[str, Any]): A dictionary of entries to extract.
        """
        kwargs = {}
        for namespaced_key, value in values.items():
            splits = namespaced_key.split("/")
            if len(splits) == 2:
                namespace, key = splits
                if namespace == cls.namespace() and key in model_fields(cls):
                    kwargs[key] = cls._extract_value(field_name=key, value=value)
                elif namespace == cls.namespace() and key in cls.current_key_by_legacy_key():
                    current_key = cls.current_key_by_legacy_key()[key]
                    if f"{cls.namespace()}/{current_key}" not in values:
                        # Only extract the value from the backcompat key if the new
                        # key is not present
                        kwargs[current_key] = cls._extract_value(
                            field_name=current_key, value=value
                        )

        return cls(**kwargs)

    @classmethod
    @abstractmethod
    def _extract_value(cls, field_name: str, value: Any) -> Any:
        """Based on type annotation, potentially coerce the value to the expected type."""
        ...

    @classmethod
    def current_key_by_legacy_key(cls) -> Mapping[str, str]:
        """Return a mapping of each legacy key to its current key."""
        return {}


class NamespacedMetadataSet(NamespacedKVSet):
    """Extend this class to define a set of metadata fields in the same namespace.

    Supports splatting to a dictionary that can be placed inside a metadata argument along with
    other dictionary-structured metadata.

    .. code-block:: python

        my_metadata: NamespacedMetadataSet = ...
        return MaterializeResult(metadata={**my_metadata, ...})
    """

    def __init__(self, *args, **kwargs) -> None:
        for field_name in model_fields(self).keys():
            annotation_types = self._get_accepted_types_for_field(field_name)
            invalid_annotation_types = {
                annotation_type
                for annotation_type in annotation_types
                if not is_raw_metadata_type(annotation_type)
            }
            if invalid_annotation_types:
                check.failed(
                    f"Type annotation for field '{field_name}' includes invalid metadata type(s): {invalid_annotation_types}"
                )
        super().__init__(*args, **kwargs)

    @classmethod
    def _extract_value(cls, field_name: str, value: Any) -> Any:
        """Based on type annotation, potentially coerce the metadata value to its inner value.

        E.g. if the annotation is Optional[float] and the value is FloatMetadataValue, construct
        the MetadataSet using the inner float.
        """
        if isinstance(value, MetadataValue):
            annotation = model_fields(cls)[field_name].annotation
            annotation_acceptable_types = flatten_unions(annotation)
            if (
                type(value) not in annotation_acceptable_types
                and type(value.value) in annotation_acceptable_types
            ):
                check.invariant(type(value.value) in DIRECTLY_WRAPPED_METADATA_TYPES)
                return value.value

        return value

    @classmethod
    @cache  # this avoids wastefully recomputing this once per instance
    def _get_accepted_types_for_field(cls, field_name: str) -> AbstractSet[type]:
        annotation = model_fields(cls)[field_name].annotation
        return flatten_unions(annotation)


class TableMetadataSet(NamespacedMetadataSet):
    """Metadata entries that apply to definitions, observations, or materializations of assets that
    are tables.

    Args:
        column_schema (Optional[TableSchema]): The schema of the columns in the table.
        column_lineage (Optional[TableColumnLineage]): The lineage of column inputs to column
            outputs for the table.
        row_count (Optional[int]): The number of rows in the table.
        partition_row_count (Optional[int]): The number of rows in the materialized or observed partition.
        table_name (Optional[str]): A unique identifier for the table/view, typically fully qualified.
            For example, `my_database.my_schema.my_table`.
    """

    column_schema: Optional[TableSchema] = None
    column_lineage: Optional[TableColumnLineage] = None
    row_count: Optional[int] = None
    partition_row_count: Optional[int] = None
    table_name: Optional[str] = None

    @classmethod
    def namespace(cls) -> str:
        return "dagster"

    @classmethod
    def current_key_by_legacy_key(cls) -> Mapping[str, str]:
        return {"relation_identifier": "table_name"}


class UriMetadataSet(NamespacedMetadataSet):
    """Metadata entry which supplies a URI address for an asset.
    For example, the S3 address of a file or bucket.

    Args:
        uri (Optional[str]): The URI address for the asset.
    """

    uri: Optional[str] = None

    @classmethod
    def namespace(cls) -> str:
        return "dagster"



from abc import ABC, abstractmethod
from enum import Enum
from typing import TYPE_CHECKING, NamedTuple, Optional

from dagster._core.instance import MayHaveInstanceWeakref, T_DagsterInstance
from dagster._core.origin import JobPythonOrigin
from dagster._core.storage.dagster_run import DagsterRun
from dagster._serdes import whitelist_for_serdes

if TYPE_CHECKING:
    from dagster._core.workspace.context import BaseWorkspaceRequestContext


class LaunchRunContext(NamedTuple):
    """Context available within a run launcher's launch_run call."""

    dagster_run: DagsterRun
    workspace: Optional["BaseWorkspaceRequestContext"]

    @property
    def job_code_origin(self) -> Optional[JobPythonOrigin]:
        return self.dagster_run.job_code_origin


class ResumeRunContext(NamedTuple):
    """Context available within a run launcher's resume_run call."""

    dagster_run: DagsterRun
    workspace: Optional["BaseWorkspaceRequestContext"]
    resume_attempt_number: Optional[int] = None

    @property
    def job_code_origin(self) -> Optional[JobPythonOrigin]:
        return self.dagster_run.job_code_origin


@whitelist_for_serdes
class WorkerStatus(Enum):
    RUNNING = "RUNNING"
    NOT_FOUND = "NOT_FOUND"
    FAILED = "FAILED"
    SUCCESS = "SUCCESS"
    UNKNOWN = "UNKNOWN"


class CheckRunHealthResult(NamedTuple):
    """Result of a check_run_worker_health call."""

    status: WorkerStatus
    msg: Optional[str] = None
    transient: Optional[bool] = None
    run_worker_id: Optional[str] = None  # Identifier for a particular run worker

    def __str__(self) -> str:
        return f"{self.status.value}: '{self.msg}'"


class RunLauncher(ABC, MayHaveInstanceWeakref[T_DagsterInstance]):
    @abstractmethod
    def launch_run(self, context: LaunchRunContext) -> None:
        """Launch a run.

        This method should begin the execution of the specified run, and may emit engine events.
        Runs should be created in the instance (e.g., by calling
        ``DagsterInstance.create_run()``) *before* this method is called, and
        should be in the ``PipelineRunStatus.STARTING`` state. Typically, this method will
        not be invoked directly, but should be invoked through ``DagsterInstance.launch_run()``.

        Args:
            context (LaunchRunContext): information about the launch - every run launcher
            will need the PipelineRun, and some run launchers may need information from the
            BaseWorkspaceRequestContext from which the run was launched.
        """

    @abstractmethod
    def terminate(self, run_id: str) -> bool:
        """Terminates a process.

        Returns False is the process was already terminated. Returns true if
        the process was alive and was successfully terminated
        """

    def dispose(self) -> None:
        """Do any resource cleanup that should happen when the DagsterInstance is
        cleaning itself up.
        """

    def join(self, timeout: int = 30) -> None:
        pass

    @property
    def supports_check_run_worker_health(self) -> bool:
        """Whether the run launcher supports check_run_worker_health."""
        return False

    def check_run_worker_health(self, run: DagsterRun) -> CheckRunHealthResult:
        raise NotImplementedError(
            "This run launcher does not support run monitoring. Please disable it on your instance."
        )

    def get_run_worker_debug_info(
        self, run: DagsterRun, include_container_logs: Optional[bool] = True
    ) -> Optional[str]:
        return None

    @property
    def supports_resume_run(self) -> bool:
        """Whether the run launcher supports resume_run."""
        return False

    def resume_run(self, context: ResumeRunContext) -> None:
        raise NotImplementedError(
            "This run launcher does not support resuming runs. If using "
            "run monitoring, set max_resume_run_attempts to 0."
        )



from typing import Any, Optional

from dagster import InputContext, OutputContext
from dagster._config.pythonic_config import ConfigurableIOManager, ResourceDependency
from dagster._core.definitions.events import AssetKey, AssetMaterialization
from dagster._core.definitions.metadata import TextMetadataValue
from dagster._core.event_api import AssetRecordsFilter
from dagster._core.events.log import EventLogEntry
from dagster._core.instance import DagsterInstance
from dagster._core.storage.io_manager import IOManager


def get_text_metadata_value(materialization: AssetMaterialization, key: str) -> Optional[str]:
    metadata_value = materialization.metadata.get(key)
    return metadata_value.value if isinstance(metadata_value, TextMetadataValue) else None


def latest_materialization_log_entry(
    instance: DagsterInstance, asset_key: AssetKey, partition_key: Optional[str] = None
) -> Optional[EventLogEntry]:
    event_records = instance.fetch_materializations(
        AssetRecordsFilter(
            asset_key=asset_key,
            asset_partitions=[partition_key] if partition_key else None,
        ),
        limit=1,
    ).records
    return event_records[0].event_log_entry if event_records else None


class BranchingIOManager(ConfigurableIOManager):
    """A branching I/O manager composes two I/O managers.

    1) The parent I/O manager, typically your production environment.
    2) The branch I/O manager, typically a development or branched environment.

    The objective of this to allow a developer to safely read from a production
    environment and then write to a separate development environment. Once data
    has been written to the branch environment subsequent reads of that asset
    are sourced from the branch environment. This bookkeeping is done in Dagster's
    asset catalog by emitting AssetMaterializations with metadata.

    This is designed for iterative development on asset graphs, especially
    where assets early in the graph are large and expensive to compute. One can
    iteratively develop on downstream assets in that graph safely.

    Some storage systems branching functionality natively. Examples include Snowflake's
    CLONE feature. Branching I/O managers allow users to implement that functionality
    in more flexible software layer over arbitrary storage systems.
    """

    parent_io_manager: ResourceDependency[IOManager]
    branch_io_manager: ResourceDependency[IOManager]

    branch_name: str = "dev"
    branch_metadata_key: str = "io_manager_branch"

    def load_input(self, context: InputContext) -> Any:
        if not context.has_asset_key:
            # we are dealing with an op input
            # just load it with the branch manager
            return self.branch_io_manager.load_input(context)
        else:
            # we are dealing with an asset input
            # figure out which partition keys are loaded, if any
            partition_keys = []
            if context.has_asset_partitions:
                partition_keys = context.asset_partition_keys

            # we'll fetch materializations with key=None if we aren't loading
            # a partitioned asset, this will return us the latest materialization
            # of an unpartitioned asset
            if len(partition_keys) == 0:
                partition_keys = [None]

            # grab the latest materialization for each partition that we
            # need to load, OR just the latest materialization if not partitioned
            event_log_entries = [
                latest_materialization_log_entry(
                    instance=context.instance,
                    asset_key=context.asset_key,
                    partition_key=partition_key,
                )
                for partition_key in partition_keys
            ]

            # if all partitions are available in the branch, we can load from the branch
            # otherwise we need to load from the parent
            if all(
                event_log_entry is not None
                and event_log_entry.asset_materialization
                and get_text_metadata_value(
                    event_log_entry.asset_materialization, self.branch_metadata_key
                )
                == self.branch_name
                for event_log_entry in event_log_entries
            ):
                context.log.info(
                    f'Branching Manager: Loading "{context.asset_key.to_user_string()}" from'
                    f' "{self.branch_name}"'
                )
                return self.branch_io_manager.load_input(context)

            context.log.info(
                f'Branching Manager Loading "{context.asset_key.to_user_string()}" from parent'
            )
            return self.parent_io_manager.load_input(context)

    def handle_output(self, context: OutputContext, obj: Any) -> None:
        # always write to the branch manager
        self.branch_io_manager.handle_output(context, obj)

        if context.has_asset_key:
            # we are dealing with an asset output (not an op output)
            # mark the asset materialization with the branch name
            context.add_output_metadata({self.branch_metadata_key: self.branch_name})
            context.log.info(
                f'Branching Manager: Writing "{context.asset_key.to_user_string()}" to branch'
                f' "{self.branch_name}"'
            )



from collections.abc import Mapping
from typing import Any, Optional

from typing_extensions import Self

import dagster._check as check
from dagster._config.config_schema import UserConfigSchema
from dagster._core.execution.api import execute_run
from dagster._core.launcher import LaunchRunContext, RunLauncher
from dagster._serdes import ConfigurableClass
from dagster._serdes.config_class import ConfigurableClassData
from dagster._utils.hosted_user_process import recon_job_from_origin


class SyncInMemoryRunLauncher(RunLauncher, ConfigurableClass):
    """This run launcher launches runs synchronously, in memory, and is intended only for test.

    Use the :py:class:`dagster.DefaultRunLauncher`.
    """

    def __init__(self, inst_data: Optional[ConfigurableClassData] = None):
        self._inst_data = inst_data
        self._repository = None
        self._instance_ref = None

        super().__init__()

    @property
    def inst_data(self) -> Optional[ConfigurableClassData]:
        return self._inst_data

    @classmethod
    def config_type(cls) -> UserConfigSchema:
        return {}

    @classmethod
    def from_config_value(
        cls, inst_data: ConfigurableClassData, config_value: Mapping[str, Any]
    ) -> Self:
        return cls(inst_data=inst_data)

    def launch_run(self, context: LaunchRunContext) -> None:
        recon_job = recon_job_from_origin(context.job_code_origin)  # type: ignore
        execute_run(recon_job, context.dagster_run, self._instance)

    def terminate(self, run_id):
        check.not_implemented("Termination not supported.")



import time
from collections import defaultdict
from types import TracebackType
from typing import TYPE_CHECKING, Optional

from typing_extensions import Self

from dagster._core.instance import DagsterInstance
from dagster._core.instance.config import PoolGranularity
from dagster._core.storage.dagster_run import DagsterRun
from dagster._core.storage.tags import PRIORITY_TAG

if TYPE_CHECKING:
    from dagster._core.storage.event_log.base import PoolLimit

INITIAL_INTERVAL_VALUE = 1
STEP_UP_BASE = 1.1
MAX_CONCURRENCY_CLAIM_BLOCKED_INTERVAL = 15

MAX_ALLOWED_PRIORITY = 2**31 - 1


class InstanceConcurrencyContext:
    """This class is used to manage instance-scoped concurrency for a given run. It wraps the
    instance-based storage methods that obtains / releases concurrency slots, and provides a common
    interface for the active execution to throttle queries to the DB to check for available slots.

    It ensures that pending concurrency claims are freed upon exiting context.  It does not,
    however, free active slots that have been claimed. This is because the executor (depending on
    the executor type) may have launched processes that may continue to run even after the current
    context is exited.

    These active slots may be manually freed via the UI, which calls the event log storage method:
    `free_concurrency_slots_by_run_id`
    """

    def __init__(self, instance: DagsterInstance, dagster_run: DagsterRun):
        self._instance = instance
        self._run_id = dagster_run.run_id
        self._pools: Optional[dict[str, PoolLimit]] = None
        self._pending_timeouts = defaultdict(float)
        self._pending_claim_counts = defaultdict(int)
        self._pending_claims = set()
        self._pool_config = instance.event_log_storage.get_pool_config()
        self._claims = set()
        try:
            self._run_priority = int(dagster_run.tags.get(PRIORITY_TAG, "0"))
        except ValueError:
            self._run_priority = 0

    def __enter__(self) -> Self:
        self._context_guard = True
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        to_clear = []
        for step_key in self._pending_claims:
            self._instance.event_log_storage.free_concurrency_slot_for_step(self._run_id, step_key)
            to_clear.append(step_key)

        for step_key in to_clear:
            if step_key in self._pending_timeouts:
                del self._pending_timeouts[step_key]
            if step_key in self._pending_claim_counts:
                del self._pending_claim_counts[step_key]

            self._pending_claims.remove(step_key)

        self._context_guard = False

    def get_pool_info(self, pool_name: str) -> Optional["PoolLimit"]:
        if not self._instance.event_log_storage.supports_global_concurrency_limits:
            return None

        if self._pools is None:
            self._sync_pools()

        assert self._pools is not None
        return self._pools.get(pool_name)

    def _sync_pools(self) -> None:
        pool_limits = self._instance.event_log_storage.get_pool_limits()
        self._pools = {pool.name: pool for pool in pool_limits}

    def claim(
        self,
        concurrency_key: str,
        step_key: str,
        step_priority: int = 0,
    ) -> bool:
        if not self._instance.event_log_storage.supports_global_concurrency_limits:
            return True

        if self._pool_config.pool_granularity == PoolGranularity.RUN:
            # with pool granularity set to run, we don't need to enforce the global concurrency
            # limits here, since we're already in a launched run.
            return True

        if step_key in self._pending_claims:
            if time.time() > self._pending_timeouts[step_key]:
                del self._pending_timeouts[step_key]
                self._sync_pools()
            else:
                return False
        else:
            self._pending_claims.add(step_key)

        default_limit = self._pool_config.default_pool_limit
        pool_info = self.get_pool_info(concurrency_key)
        if (pool_info is None and default_limit is not None) or (
            pool_info is not None and pool_info.from_default and pool_info.limit != default_limit
        ):
            self._instance.event_log_storage.initialize_concurrency_limit_to_default(
                concurrency_key
            )
            self._sync_pools()
            # refetch the pool info
            pool_info = self.get_pool_info(concurrency_key)

        if pool_info is None:
            # make sure we remove claims here, since we could have previously blocked on a pool,
            # which then had its limit removed
            if step_key in self._pending_claims:
                self._pending_claims.remove(step_key)
            return True

        priority = self._run_priority + step_priority

        if abs(priority) > MAX_ALLOWED_PRIORITY:
            raise Exception(
                f"Tried to claim a concurrency slot with a priority {priority} that was not in the allowed range of a 32-bit signed integer."
            )

        claim_status = self._instance.event_log_storage.claim_concurrency_slot(
            concurrency_key, self._run_id, step_key, priority
        )

        if not claim_status.is_claimed:
            interval = _calculate_timeout_interval(
                claim_status.sleep_interval, self._pending_claim_counts[step_key]
            )
            self._pending_timeouts[step_key] = time.time() + interval
            self._pending_claim_counts[step_key] += 1
            return False

        if step_key in self._pending_claims:
            self._pending_claims.remove(step_key)

        self._claims.add(step_key)
        return True

    def interval_to_next_pending_claim_check(self) -> float:
        if not self._pending_claims:
            return 0.0

        now = time.time()
        return min([0, *[ready_at - now for ready_at in self._pending_timeouts.values()]])

    def pending_claim_steps(self) -> list[str]:
        return list(self._pending_claims)

    def has_pending_claims(self) -> bool:
        return len(self._pending_claims) > 0

    def free_step(self, step_key) -> None:
        if step_key not in self._claims:
            return

        self._instance.event_log_storage.free_concurrency_slot_for_step(self._run_id, step_key)
        self._claims.remove(step_key)


def _calculate_timeout_interval(sleep_interval: Optional[float], pending_claim_count: int) -> float:
    if sleep_interval is not None:
        return sleep_interval

    if pending_claim_count > 30:
        # with the current values, we will always hit the max by the 30th claim attempt
        return MAX_CONCURRENCY_CLAIM_BLOCKED_INTERVAL

    # increase the step up value exponentially, up to a max of 15 seconds (starting from 0)
    step_up_value = STEP_UP_BASE**pending_claim_count - 1
    interval = INITIAL_INTERVAL_VALUE + step_up_value
    return min(MAX_CONCURRENCY_CLAIM_BLOCKED_INTERVAL, interval)



from abc import ABC, abstractmethod
from collections.abc import Iterable
from functools import partial
from typing import TYPE_CHECKING, Generic, Optional, TypeVar

from typing_extensions import Self

import dagster._check as check
from dagster._utils.aiodataloader import BlockingDataLoader, DataLoader

TResult = TypeVar("TResult")
TKey = TypeVar("TKey")


if TYPE_CHECKING:
    from dagster._core.instance import DagsterInstance

"""
Loadable - asyncio driven batching

    When resolving a GraphQL operation, it is common to end up with a pattern where a
    set of objects are resolved, and then a field on those objects requires fetching
    a related object by id. To avoid serially fetching those objects by id, something
    must be done. In this past, this batching was solved by manual bespoke classes.

    Loadable is an abstraction that replaces this explicit batching with an implicit scheme
    leveraging the asyncio event loop.

    Within a LoadingContext, requests for the same Loadable object are routed through a shared
    DataLoader instance. This instance accumulates requests that happen in the same event loop
    pass and batches them together. Any subsequent requests for the same ID in the same
    LoadingContext will be served from an in memory cache. This means that mutations need to
    clear the LoadingContext cache.

    The Loadable interface is two methods, `gen` for fetching a single object by its id
    and `gen_many` for loading N by their ids. The "gen" (for generator)[1] name is chosen
    to stand out and make it clear that the method needs to be awaited, which at this time
    is anomalous in a codebase where async is rarely used.

Additional resources:
* https://xuorig.medium.com/the-graphql-dataloader-pattern-visualized-3064a00f319f
[1] Brought to you by the 201X Facebook codebase
"""


class LoadingContext(ABC):
    """A scoped object in which Loadable objects will be fetched in batches and cached.

    Expected to be implemented by request scoped context objects that have access to the DagsterInstance.
    """

    @property
    @abstractmethod
    def instance(self) -> "DagsterInstance":
        raise NotImplementedError()

    @property
    @abstractmethod
    def loaders(self) -> dict[type, tuple[DataLoader, BlockingDataLoader]]:
        raise NotImplementedError()

    def get_loaders_for(self, ttype: type["LoadableBy"]) -> tuple[DataLoader, BlockingDataLoader]:
        if ttype not in self.loaders:
            if not issubclass(ttype, LoadableBy):
                check.failed(f"{ttype} is not Loadable")

            batch_load_fn = partial(ttype._batch_load, context=self)  # noqa
            blocking_batch_load_fn = partial(ttype._blocking_batch_load, context=self)  # noqa

            self.loaders[ttype] = (
                DataLoader(batch_load_fn=batch_load_fn),
                BlockingDataLoader(batch_load_fn=blocking_batch_load_fn),
            )

        return self.loaders[ttype]

    def clear_loaders(self) -> None:
        for ttype in self.loaders:
            del self.loaders[ttype]


# Expected there may be other "Loadable" base classes based on what is needed to load.


class LoadableBy(ABC, Generic[TKey]):
    """Make An object Loadable by ID of type TKey using a DagsterInstance."""

    @classmethod
    async def _batch_load(
        cls, keys: Iterable[TKey], context: "LoadingContext"
    ) -> Iterable[Optional[Self]]:
        return cls._blocking_batch_load(keys, context)

    @classmethod
    @abstractmethod
    def _blocking_batch_load(
        cls, keys: Iterable[TKey], context: "LoadingContext"
    ) -> Iterable[Optional[Self]]:
        # There is no good way of turning an async function into a sync one that
        # will allow us to execute that sync function inside of a broader async context.
        #
        # In the spirit of allowing incremental migration from a fully-sync pattern to
        # an async one, we provide two separate functions here to allow sync and async
        # calls to coexist.
        raise NotImplementedError()

    @classmethod
    async def gen(cls, context: LoadingContext, id: TKey) -> Optional[Self]:
        """Fetch an object by its id."""
        loader, _ = context.get_loaders_for(cls)
        return await loader.load(id)

    @classmethod
    async def gen_many(
        cls, context: LoadingContext, ids: Iterable[TKey]
    ) -> Iterable[Optional[Self]]:
        """Fetch N objects by their id."""
        loader, _ = context.get_loaders_for(cls)
        return await loader.load_many(ids)

    @classmethod
    def blocking_get(cls, context: LoadingContext, id: TKey) -> Optional[Self]:
        """Fetch an object by its id."""
        _, blocking_loader = context.get_loaders_for(cls)
        return blocking_loader.blocking_load(id)

    @classmethod
    def blocking_get_many(cls, context: LoadingContext, ids: Iterable[TKey]) -> Iterable[Self]:
        """Fetch N objects by their id."""
        # in the future, can consider priming the non-blocking loader with the results of this
        # sync call
        _, blocking_loader = context.get_loaders_for(cls)
        return list(filter(None, blocking_loader.blocking_load_many(ids)))

    @classmethod
    def prepare(cls, context: LoadingContext, ids: Iterable[TKey]) -> None:
        """Ensure the provided ids will be fetched on the next blocking query."""
        _, blocking_loader = context.get_loaders_for(cls)
        blocking_loader.prepare(ids)


class LoadingContextForTest(LoadingContext):
    """Loading context intended to be used in unit tests that would not otherwise construct a LoadingContext."""

    def __init__(self, instance: "DagsterInstance"):
        self._instance = instance
        self._loaders = {}

    @property
    def instance(self) -> "DagsterInstance":
        return self._instance

    @property
    def loaders(self) -> dict[type, tuple[DataLoader, BlockingDataLoader]]:
        return self._loaders



from collections.abc import Mapping, Sequence
from typing import Any, Optional, cast

import dagster._check as check
from dagster._config.config_type import ConfigScalarKind, ConfigType, ConfigTypeKind
from dagster._config.field import Field
from dagster._record import IHaveNew, record, record_custom
from dagster._serdes import whitelist_for_serdes


def get_recursive_type_keys(
    config_type_snap: "ConfigTypeSnap", config_schema_snapshot: "ConfigSchemaSnapshot"
) -> set[str]:
    check.inst_param(config_type_snap, "config_type_snap", ConfigTypeSnap)
    check.inst_param(config_schema_snapshot, "config_schema_snapshot", ConfigSchemaSnapshot)
    result_keys = set()
    for type_key in config_type_snap.get_child_type_keys():
        result_keys.add(type_key)
        for recurse_key in get_recursive_type_keys(
            config_schema_snapshot.get_config_snap(type_key), config_schema_snapshot
        ):
            result_keys.add(recurse_key)
    return result_keys


def _ct_name(ct):
    return ct.name


@whitelist_for_serdes(skip_when_empty_fields={"field_aliases"})
@record_custom
class ConfigTypeSnap(IHaveNew):
    kind: ConfigTypeKind
    key: str
    given_name: Optional[str]
    description: Optional[str]
    type_param_keys: Optional[Sequence[str]]  # only valid for closed generics
    enum_values: Optional[Sequence["ConfigEnumValueSnap"]]  # only valid for enums
    fields: Optional[Sequence["ConfigFieldSnap"]]  # only valid for dicts and selectors
    scalar_kind: Optional[ConfigScalarKind]  # only valid for scalars
    field_aliases: Optional[Mapping[str, str]]  # only valid for strict shapes

    # serdes log
    # * Adding scalar_kind
    # * Adding field_aliases
    def __new__(
        cls,
        kind: ConfigTypeKind,
        key: str,
        given_name: Optional[str],
        description: Optional[str],
        type_param_keys: Optional[Sequence[str]],
        enum_values: Optional[Sequence["ConfigEnumValueSnap"]],
        fields: Optional[Sequence["ConfigFieldSnap"]],
        # Old version of object will not have these properties
        scalar_kind: Optional[ConfigScalarKind] = None,
        field_aliases: Optional[Mapping[str, str]] = None,
    ):
        return super().__new__(
            cls,
            kind=kind,
            key=key,
            given_name=given_name,
            description=description,
            type_param_keys=type_param_keys,
            enum_values=enum_values,
            fields=(
                None
                if fields is None
                else sorted(
                    check.list_param(fields, "field", of_type=ConfigFieldSnap),
                    key=_ct_name,
                )
            ),
            scalar_kind=scalar_kind,
            field_aliases=field_aliases or {},
        )

    @property
    def key_type_key(self) -> str:
        """For a type which has keys such as Map, returns the type of the key."""
        # valid for Map, which has its key type as the first entry in type_param_keys
        check.invariant(self.kind == ConfigTypeKind.MAP)

        type_param_keys = check.is_list(self.type_param_keys, of_type=str)
        check.invariant(len(type_param_keys) == 2)
        return type_param_keys[0]

    @property
    def inner_type_key(self) -> str:
        """For container types such as Array or Noneable, the contained type. For a Map, the value type."""
        # valid for Noneable, Map, and Array
        check.invariant(
            self.kind == ConfigTypeKind.NONEABLE
            or self.kind == ConfigTypeKind.ARRAY
            or self.kind == ConfigTypeKind.MAP
        )

        type_param_keys = check.is_list(self.type_param_keys, of_type=str)
        if self.kind == ConfigTypeKind.MAP:
            # For a Map, the inner (value) type is the second entry (the first is the key type)
            check.invariant(len(type_param_keys) == 2)
            return type_param_keys[1]
        else:
            check.invariant(len(type_param_keys) == 1)
            return type_param_keys[0]

    @property
    def scalar_type_key(self) -> str:
        check.invariant(self.kind == ConfigTypeKind.SCALAR_UNION)
        type_param_keys = check.is_list(self.type_param_keys, of_type=str)
        return type_param_keys[0]

    @property
    def non_scalar_type_key(self) -> str:
        check.invariant(self.kind == ConfigTypeKind.SCALAR_UNION)
        type_param_keys = check.is_list(self.type_param_keys, of_type=str)
        return type_param_keys[1]

    def _get_field(self, name: str) -> Optional["ConfigFieldSnap"]:
        check.str_param(name, "name")
        check.invariant(ConfigTypeKind.has_fields(self.kind))
        fields = check.is_list(self.fields, of_type=ConfigFieldSnap)
        for f in fields:
            if f.name == name:
                return f

        return None

    def get_field(self, name: str) -> "ConfigFieldSnap":
        field = self._get_field(name)
        if not field:
            check.failed(f"Field {name} not found")
        return field

    def has_field(self, name: str) -> bool:
        return bool(self._get_field(name))

    @property
    def field_names(self) -> Sequence[str]:
        fields = check.is_list(self.fields, of_type=ConfigFieldSnap)
        # Typing error caught by making is_list typed -- schrockn 2024-06-09
        return [fs.name for fs in fields]  # type: ignore

    def get_child_type_keys(self) -> Sequence[str]:
        if ConfigTypeKind.is_closed_generic(self.kind):
            # all closed generics have type params
            return cast(list[str], self.type_param_keys)
        elif ConfigTypeKind.has_fields(self.kind):
            return [
                field.type_key for field in cast(list[ConfigFieldSnap], check.not_none(self.fields))
            ]
        else:
            return []

    def has_enum_value(self, value: object) -> bool:
        check.invariant(self.kind == ConfigTypeKind.ENUM)
        for enum_value in cast(list[ConfigEnumValueSnap], self.enum_values):
            if enum_value.value == value:
                return True
        return False


@whitelist_for_serdes
@record
class ConfigEnumValueSnap:
    value: str
    description: Optional[str]


@whitelist_for_serdes
@record
class ConfigFieldSnap:
    name: Optional[str]
    type_key: str
    is_required: bool
    default_provided: bool
    default_value_as_json_str: Optional[str]
    description: Optional[str]


def snap_from_field(name: str, field: Field):
    return ConfigFieldSnap(
        name=name,
        type_key=field.config_type.key,
        is_required=field.is_required,
        default_provided=field.default_provided,
        default_value_as_json_str=(
            field.default_value_as_json_str if field.default_provided else None
        ),
        description=field.description,
    )


# type-ignores here are temporary until config type system overhauled
def snap_from_config_type(config_type: ConfigType) -> ConfigTypeSnap:
    return ConfigTypeSnap(
        key=config_type.key,
        given_name=config_type.given_name,
        kind=config_type.kind,
        description=config_type.description,
        type_param_keys=(
            [ct.key for ct in config_type.type_params]
            if config_type.type_params
            # jam scalar union types into type_param_keys
            else (
                [config_type.scalar_type.key, config_type.non_scalar_type.key]  # type: ignore
                if config_type.kind == ConfigTypeKind.SCALAR_UNION
                else None
            )
        ),
        enum_values=(
            [
                ConfigEnumValueSnap(value=ev.config_value, description=ev.description)
                for ev in config_type.enum_values  # type: ignore
            ]
            if config_type.kind == ConfigTypeKind.ENUM
            else None
        ),
        fields=(
            [snap_from_field(name, field) for name, field in config_type.fields.items()]  # type: ignore
            if ConfigTypeKind.has_fields(config_type.kind)
            else None
        ),
        scalar_kind=config_type.scalar_kind if config_type.kind == ConfigTypeKind.SCALAR else None,  # type: ignore
        field_aliases=(
            config_type.field_aliases  # type: ignore
            if config_type.kind == ConfigTypeKind.STRICT_SHAPE
            else None
        ),
    )


@whitelist_for_serdes
@record
class ConfigSchemaSnapshot:
    all_config_snaps_by_key: Mapping[str, ConfigTypeSnap]

    @property
    def all_config_keys(self) -> Sequence[str]:
        return list(self.all_config_snaps_by_key.keys())

    def get_config_snap(self, key: str) -> ConfigTypeSnap:
        check.str_param(key, "key")
        return self.all_config_snaps_by_key[key]

    def has_config_snap(self, key: str) -> bool:
        check.str_param(key, "key")
        return key in self.all_config_snaps_by_key


def minimal_config_for_type_snap(
    config_schema_snap: ConfigSchemaSnapshot, config_type_snap: ConfigTypeSnap
) -> Any:
    check.inst_param(config_schema_snap, "config_schema_snap", ConfigSchemaSnapshot)
    check.inst_param(config_type_snap, "config_type_snap", ConfigTypeSnap)

    if ConfigTypeKind.has_fields(config_type_snap.kind):
        default_dict = {}
        if ConfigTypeKind.is_selector(config_type_snap.kind):
            return "<selector>"
        for field in config_type_snap.fields:  # type: ignore
            if not field.is_required:
                continue

            default_dict[field.name] = minimal_config_for_type_snap(
                config_schema_snap, config_schema_snap.get_config_snap(field.type_key)
            )
        return default_dict
    elif config_type_snap.kind == ConfigTypeKind.ANY:
        return "AnyType"
    elif config_type_snap.kind == ConfigTypeKind.SCALAR:
        defaults = {"String": "...", "Int": 0, "Float": 0.0, "Bool": True}

        return defaults.get(config_type_snap.given_name, "<unknown>")  # type: ignore
    elif config_type_snap.kind == ConfigTypeKind.ARRAY:
        return []
    elif config_type_snap.kind == ConfigTypeKind.MAP:
        return {}
    elif config_type_snap.kind == ConfigTypeKind.ENUM:
        # guard against the edge case that an enum is defined with zero options
        return (
            config_type_snap.enum_values[0].value if config_type_snap.enum_values else "<unknown>"
        )
    elif config_type_snap.kind == ConfigTypeKind.SCALAR_UNION:
        return minimal_config_for_type_snap(
            config_schema_snap,
            config_schema_snap.get_config_snap(config_type_snap.type_param_keys[0]),  # type: ignore
        )
    else:
        return "<unknown>"
